"""The LinxProcessor for digital lock-in."""

import logging
import os
from pathlib import Path
from typing import Literal, Self

import numpy as np
from scipy import optimize, signal

from ..core import BaseProcessor, sp
from ..log import configure_logger
from ._config import LinxConfig

PICKUP_NAME = "pickup"
SERIE_NAME = ""

logger = logging.getLogger(__name__)


class LinxProcessor(BaseProcessor):
    """A Processor class for digital lock-in."""

    cfg: LinxConfig

    def __init__(self, *args, **kwargs):
        configure_logger(logger, "linx.log")
        self._config_cls = LinxConfig

        # Serie name for the NeXus data processed group name
        self._serie_name = SERIE_NAME

        super().__init__(*args, **kwargs)

        # Initialize attributes
        if not self.is_nexus_file:
            self._prepare_metadata()
        else:
            # Fill missing attributes
            self.npoints = self.get_data_processed("magfield").size
            self.nchannels = len(self.cfg.measurements)
            self._bin_data = np.array([])

        # Store names
        self._pu_name = PICKUP_NAME
        # Read signal and reference available names
        self._signal_names = self._get_signal_names()
        # Check those names are set correctly in the settings
        self.ref_name = self.cfg.settings.ref_name
        self.signal_name = self.cfg.settings.signal_name

    @property
    def idx_serie(self) -> str:
        """
        Track the current signal.

        The returned value is read from the "signal_name" parameter in the "settings"
        section of the configuration.
        """
        return self.cfg.settings.signal_name

    @idx_serie.setter
    def idx_serie(self, value: str):
        """
        Setter for `idx_serie`.

        The "signal_name" parameter in the "settings" section of the configuration is
        updated and the NeXus group for this serie is created.
        """
        if value not in self._signal_names:
            raise ValueError(
                f"'{value}' was not found in the measurements, check your settings"
            )
        self.cfg.settings.signal_name = value
        self.metadata["phase_offset"][value] = self.cfg.settings.phase_offset
        self.create_data_serie()

    @property
    def signal_name(self) -> str:
        """
        Track the current signal. Alias for `idx_serie`.

        The returned value is read from the "signal_name" parameter in the "settings"
        section of the configuration.
        """
        return self.idx_serie

    @signal_name.setter
    def signal_name(self, value: str):
        """Setter for `signal_name`, the alias for `idx_serie`."""
        self.idx_serie = value
        logger.info(f"Using signal : '{value}'")

    @property
    def ref_name(self) -> str:
        """
        Track the current reference.

        The returned value is read from the "ref_name" parameter in the "settings"
        section of the configuration.
        """
        return self.cfg.settings.ref_name

    @ref_name.setter
    def ref_name(self, value: str):
        """
        Setter for `ref_name`.

        The "ref_name" parameter in the "settings" section of the configuration is
        updated.
        """
        if value not in self._signal_names:
            raise ValueError(
                f"'{value}' was not found in the measurements, check your settings"
            )
        self.cfg.settings.ref_name = value
        logger.info(f"Using reference : '{value}'")

    @property
    def f0(self) -> float:
        """
        Get the center frequency of the current reference signal.

        It is read from the metadata dictionary.
        """
        if not self._check_f0_computed():
            logger.error(f"f0 was not computed for '{self.signal_name}'")
            return -1.0
        return self.metadata["f0"][self.signal_name]

    @f0.setter
    def f0(self, value: float):
        """
        Set the center frequency of the current reference signal.

        It is set in the metadata dictionary and in the configuration.
        """
        self.metadata["f0"][self.signal_name] = value
        self.cfg.settings.f0 = value

    @property
    def phase_offset(self) -> float:
        """
        Get the phase offset of the current reference signal.

        It is read from the metadata dictionary.
        """
        return self.metadata["phase_offset"][self.signal_name]

    @phase_offset.setter
    def phase_offset(self, value: float):
        """
        Set the phase offset of the current reference signal.

        It is set in the metadata dictionary, and in the configuration.
        """
        self.metadata["phase_offset"][self.signal_name] = value
        self.cfg.settings.phase_offset = value

    @property
    def amp0(self) -> float:
        """
        Get the amplitude of the current reference signal.

        It is read from the metadata dictionary.
        """
        if not self._check_demodulated():
            logger.warning(f"amp0 was not computed for '{self.ref_name}'")
            return 0
        return self.metadata["amp0"][self.signal_name]

    @amp0.setter
    def amp0(self, value: float):
        """
        Set the amplitude of the current reference signal.

        It is set in the metadata dictionary.
        """
        self.metadata["amp0"][self.signal_name] = value

    def _reinit(self) -> None:
        """Reinitialize data objects."""
        super()._reinit()

        self._prepare_metadata()

    def load_pickup(self) -> Self:
        """Read pickup time serie from the binary file."""
        if self._bin_data.size == 0:
            self._load_bin_file()

        # Get the correct index
        pickup_idx = self.cfg.measurements[self._pu_name]

        # Set data
        logger.info("Reading pickup data...")
        self.set_data_raw("pickup", self._bin_data[:, pickup_idx])
        if self.get_data_raw("time_exp", checkonly=True):
            self.set_data_raw("pickup_time", self.get_data_raw("time_exp"))
        else:
            self.set_data_raw(
                "pickup_time",
                self.get_time_exp(),
            )
        logger.info("Pickup signal loaded")

        return self

    def compute_field(self, method: str = "trapz") -> Self:
        """Compute magnetic field."""
        super().compute_field()

        # Apply time offset
        self.apply_time_offset()

        return self

    def load_data(self) -> Self:
        """Load raw data."""
        self.load_oscillo()
        return self

    def preprocess(self) -> Self:
        """Find reference center frequency f0."""
        self.find_f0()
        return self

    def align_field(self, on_pickup: bool = True, **kwargs) -> Self:
        """
        Align magnetic field on the `time_exp` vector.

        **kwargs are ignored.
        """
        if not self.get_data_processed("time_exp", checkonly=True):
            logger.warning("Experiment time vector not built, load data first")
            return self

        # Align the pickup signal
        super().align_field(on_pickup=on_pickup, left=0, right=0)

        return self

    def run_analysis(self) -> Self:
        """Run the analysis : perform the demodulation and computes the resistance."""
        self.demodulate().compute_resistance()
        return self

    def export_csv(
        self,
        fname: str | Path | None = None,
        sep: str = "\t",
        downsample: int = 0,
        average: bool = False,
    ) -> None:
        r"""
        Export results to a CSV file.

        The CSV will contain the following columns : time, pickup, B, R_in_phase, R_out_phase.

        Optionally, downsample the data with or without averaging.

        Parameters
        ----------
        fname : str or Path or None, optional
            Path to the output file. If None, the filename is generated automatically.
            Default is None.
        sep : str, optional
            Delimiter in the CSV file. Default is "\t".
        downsample : int, optional
            Downsample the data before exporting. Disable with 0 (default).
        average : bool, optional
            If downsampling is enabled, use the average instead of direct indexing.
            Default is True.
        """
        if fname is None:
            fname = self.get_csv_filename()

        logger.info(f"Saving results at {fname}...")

        # Collect data
        if self.get_data_processed("pickup", checkonly=True):
            pickup = self.get_data_processed("pickup")
        else:
            pickup = self.get_data_raw("pickup")
        data = np.concatenate(
            (
                self.get_data_processed("time_exp")[..., np.newaxis],
                pickup[..., np.newaxis],
                self.get_data_processed("magfield")[..., np.newaxis],
                self.get_data_serie("resistance_in_phase")[..., np.newaxis],
                self.get_data_serie("resistance_out_phase")[..., np.newaxis],
            ),
            axis=1,
        )
        if downsample > 0:
            data = sp.subsample_array(
                data, downsample, start=0, average=average, axis=0
            )

        # Configure output file
        header = sep.join(("time", "pickup", "field", "R_in_phase", "R_out_phase"))

        # Save
        np.savetxt(fname, data, delimiter=sep, header=header, comments="")

        logger.info("Results saved")

    def get_bin_file_properties(self) -> Self:
        """Determine binary file number columns and number of samples."""
        # Get parameters
        filepath = self.cfg.filenames["oscillo"]
        precision = self.cfg.files["oscillo"].precision

        self.nchannels, self.nsamples = self._get_bin_file_properties(
            filepath, precision=precision
        )

        return self

    def load_oscillo(self) -> Self:
        """Read the reference and signal data for the current measurement."""
        if self._bin_data.size == 0:
            self._load_bin_file()

        # Get the correct indices
        ref_idx = self.cfg.measurements[self.ref_name]
        signal_idx = self.cfg.measurements[self.signal_name]

        # Set the data
        logger.info(f"Reading {self.ref_name}...")
        self.set_data_raw(self.ref_name, self._bin_data[:, ref_idx])
        logger.info(f"Reading {self.signal_name}...")
        self.set_data_raw(self.signal_name, self._bin_data[:, signal_idx])
        if self.get_data_raw("pickup_time", checkonly=True):
            time_exp = self.get_data_raw("pickup_time")
        else:
            time_exp = self.get_time_exp()
        self.set_data_raw("time_exp", time_exp)
        # Set in data_processed as well
        self.set_data_processed("time_exp", time_exp)

        logger.info("Data loaded")

        return self

    def get_time_exp(self) -> np.ndarray:
        """Build the experiment time vector."""
        return np.linspace(
            0,
            self.nsamples / self.cfg.parameters.samplerate,
            self.nsamples,
            endpoint=False,
        )

    def apply_time_offset(self) -> Self:
        """Apply a time offset on the magnetic field time vector."""
        time_offset = self.cfg.settings.time_offset

        if not self._check_field_computed():
            logger.warning("Field not computed, cannot apply time offset")
            return self

        # Get the original time vector in case an offset was already applied
        if not self.get_data_raw("time_exp", checkonly=True):
            # Build the time vector in case it was not done before
            self.set_data_raw("time_exp", self.get_time_exp())
        time_vector = self.get_data_raw("time_exp")
        logger.info(f"Applying a {time_offset * 1e3}ms offset on magnetic field...")
        self.set_data_processed("magfield_time", time_vector - time_offset)
        logger.info("Time offset applied on magnetic field")

        # Align field if it was created
        if self.get_data_processed("time_exp", checkonly=True):
            self.align_field(on_pickup=True)

        return self

    def find_f0(self) -> Self:
        """Find the center frequency in the reference signal."""
        if not self._check_data_loaded():
            self.load_oscillo()

        logger.info(f"Finding center frequency in '{self.ref_name}'...")
        self.f0 = sp.find_f0(
            self.get_data_raw(self.ref_name), self.cfg.parameters.samplerate
        ).mean()
        logger.info(f"Found {self.f0 / 1000:3.3f}kHz")

        return self

    def build_lowpass_filter(
        self,
    ) -> (
        tuple[np.ndarray, Literal["sos", "ba"]]
        | tuple[tuple[np.ndarray, np.ndarray], Literal["sos", "ba"]]
    ):
        """Build low-pass filter."""
        filter_order = self.cfg.settings.filter_order
        filter_fc = self.cfg.settings.filter_fc
        filter_edge = self.cfg.settings.filter_edge
        fs = self.cfg.parameters.samplerate
        pb_mult = self.cfg.settings.filter_passband_mult
        sb_mult = self.cfg.settings.filter_stopband_mult
        pb_att = self.cfg.settings.filter_passband_att
        sb_att_mult = self.cfg.settings.filter_stopband_att_mult

        # Filter properties
        wp = filter_fc * pb_mult / fs  # passband
        ws = wp * sb_mult  # stopband
        gpass = pb_att  # max loss in passband
        gstop = filter_order * sb_att_mult  # min attenuation in stopband
        order, wn = signal.buttord(wp, ws, gpass, gstop)

        # Create filter
        if filter_edge == "pad":
            # Using padding to apply the filter, so we can use second order sections
            res = signal.butter(order, wn, btype="lowpass", output="sos")
            filt_type = "sos"
        elif filter_edge == "gust":
            # Using Gustafsson’s method, we need numerator/denominator
            res = signal.butter(order, wn, btype="lowpass", output="ba")
            filt_type = "ba"
        else:
            raise ValueError(
                f"{filter_edge} is not a valid method, choose 'pad' or 'gust'"
            )

        return res, filt_type

    def demodulate(self) -> Self:
        """Digital demodulation."""
        # Checks
        if not self._check_data_loaded():
            logger.warning("Data not loaded")
            return self
        if not self._check_f0_computed():
            self.find_f0()

        # Get reference phase fit parameters
        fs = self.cfg.parameters.samplerate
        ncycles = self.cfg.settings.phase_fit_ncycles
        f0 = self.f0

        # Define cos function to fit reference phase
        def cos_func(
            x: np.ndarray, A: float | np.ndarray, phi: float | np.ndarray
        ) -> np.ndarray:
            return A * np.cos(2 * np.pi * f0 * x + phi)

        # Get signals
        fit_npoints = int(ncycles * fs / f0)  # number of points to use for the fit
        reftime = self.get_data_raw("time_exp")[:fit_npoints]
        ref = self.get_data_raw(self.ref_name)[:fit_npoints]
        sigtime = self.get_data_raw("time_exp")
        sig = self.get_data_raw(self.signal_name)

        # Fit reference signal to find the phase
        logger.info(f"Fitting '{self.ref_name}' phase...")
        amp, phi = sp.fit_signal(
            cos_func, reftime, ref, bounds=([0, -np.inf], [np.inf, np.inf])
        )
        amp = float(amp)  # for type checking
        phi = float(phi)  # for type checking
        logger.info(f"Found amplitude: {amp:2.4f}V and phase: {phi:0.4f}rad")

        # Build filter
        logger.info("Building low-pass filter...")
        filt, filt_type = self.build_lowpass_filter()
        logger.info("Low-pass fitler built")

        # Demodulation
        phase_offset = self.phase_offset
        logger.info(
            f"Demodulation at {self.f0 / 1e3:2.4f}kHz (using phase: "
            f"{phi + phase_offset:0.4f}rad)..."
        )
        in_phase, out_phase = self._demodulate(
            filt_type,
            None,
            None,
            sigtime,
            sig,
            cos_func,
            filt,
            extra_phase=phase_offset,
            refas=amp,
            refphis=phi,
        )
        logger.info("Signal demodulated")

        # Store
        self.amp0 = amp
        self.set_data_serie("in_phase", in_phase.ravel())
        self.set_data_serie("out_phase", out_phase.ravel())

        # Prepare placeholders that will be replaced by actual NXlink when the
        # object has a NXroot entry
        time_exp = f"!link to:processed/analysis/{self._results_name}/time_exp"
        magfield = f"!link to:processed/analysis/{self._results_name}/magfield"
        # Store the link in the 'serie' group
        self.set_data_serie("time_exp", time_exp)
        self.set_data_serie("magfield", magfield)

        return self

    def find_phase_offset(self) -> Self:
        """Find the phase shift that maximises signal in I."""
        if not self._check_demodulated():
            logger.warning("Cannot find phase, demodulation was not performed")
            return self
        # Get parameters
        method = self.cfg.settings.auto_phase_method
        if method == "at peak field":
            if not self._check_field_computed():
                logger.warning(
                    f"Field was not computed, cannot find phase with method {method}"
                )
                return self
            idx_max_field = int(np.argmax(self.get_data_processed("magfield")))
        else:
            idx_max_field = 0

        # Get signals
        in_phase = self.get_data_serie("in_phase")
        out_phase = self.get_data_serie("out_phase")

        # Compute
        logger.info(f"Finding phase with the '{method}' method...")
        phase = self._find_phase(in_phase, out_phase, method, idx_max=idx_max_field)
        logger.info(f"Found {phase:1.3f} rad ({np.rad2deg(phase):3.3f}deg)")

        # Store
        self.phase_offset = phase
        return self

    def compute_resistance(self) -> Self:
        """Derive resistance in ohms from demodulated signal."""
        # Checks
        if not self._check_demodulated():
            logger.warning("Cannot compute resistance, demodulation was not performed")
            return self

        # Get parameters
        current = self.cfg.parameters.current
        gain = self.cfg.parameters.gain
        amp0 = self.amp0
        rms = self.cfg.parameters.gain_rms

        logger.info("Computing resistance...")
        intgain = 0.5 * amp0 * gain * current
        if rms:
            intgain *= np.sqrt(2)

        res_in = self.get_data_serie("in_phase") / intgain
        res_out = self.get_data_serie("out_phase") / intgain

        logger.info(f"Done, found max. resistance {res_in.max():2.3f}ohm")

        # Store
        self.set_data_serie("resistance_in_phase", res_in)
        self.set_data_serie("resistance_out_phase", res_out)

        return self

    def get_csv_filename(self) -> str:
        """
        Generate output CSV file name for the current signal.

        The file is placed in the data directory set in the configuration. Its name
        contains the signal name.
        """
        fout = os.path.join(self.cfg.data_directory, self.cfg.expid + "-results")

        return f"{fout}_{self.signal_name}.csv"

    def _prepare_metadata(self) -> None:
        """Prepare metadata and data objects."""
        # Metadata
        self.metadata = dict()
        self.metadata["f0"] = dict()  # will be signal_name: f0
        self.metadata["phase_offset"] = dict()  # will be signal_name: phase
        self.metadata["amp0"] = dict()  # will be signal_name: amplitude

        self.nchannels = 0
        self.nsamples = 0

        self._bin_data = np.array([])

        # Read binary file properties
        self.get_bin_file_properties()
        # Prepare data groups
        self.create_data_serie()

        # Fill metadata from configuration
        self.phase_offset = self.cfg.settings.phase_offset

    def _get_signal_names(self) -> set[str]:
        """Determine signal and reference names from configuration file measurements."""
        return {*self.cfg.measurements.keys()} - {self._pu_name}

    def _get_bin_file_properties(
        self, filepath: str | Path, precision: int = 4
    ) -> tuple[int, int]:
        """
        Determine binary file number columns and number of samples.

        Parameters
        ----------
        filepath : str or Path
            Path to the binary file.
        precision : int, optional
            Precision of floats in the binary file (typically, 4, i.e. float32). Default
            is 4.

        Returns
        -------
        nchannels : int
            Number of columns in the binary file.
        nsamples : int
            Number of data points per channel.
        """
        if isinstance(filepath, str):
            filepath = Path(filepath)

        ncols = len(self.cfg.measurements.keys())  # number of expected columns
        filesize = filepath.stat().st_size  # size in bytes
        nsamples = int(filesize / (ncols * precision))

        return ncols, nsamples

    def _load_bin_file(self) -> None:
        """
        Load binary the binary file in a temporary variable.

        Since all channels, including the pickup signal, are stored in the same binary
        file and in a not-contiguous manner, it is not possible to load a single channel
        from the file. Therefore, we cache the whole file here and access and store its
        data on-demand.
        """
        # Get parameters
        filepath = self.cfg.filenames["oscillo"]
        precision = self.cfg.files["oscillo"].precision
        endian = self.cfg.files["oscillo"].endian
        order = self.cfg.files["oscillo"].order

        # Read the time series
        self._bin_data = (
            np.fromfile(
                filepath,
                dtype=f"{endian}f{precision}",
                count=self.nchannels * self.nsamples,
            )
            .reshape((self.nsamples, self.nchannels), order=order)
            .astype(float)
        )

    def _demodulate(
        self,
        filt_type: Literal["sos", "ba"],
        reftime: np.ndarray | None,
        ref: np.ndarray | None,
        sigtime,
        sig,
        cos_func,
        filt: np.ndarray | tuple[np.ndarray, np.ndarray],
        extra_phase: float = 0.0,
        refas: float | None = None,
        refphis: float | None = None,
    ) -> tuple[np.ndarray, np.ndarray]:
        """
        Perform digital demodulation.

        Wrapper around `sp.demodulate()`, determining filter type and edge handling.
        """
        # Determine how to apply low-pass filter
        fs = self.cfg.parameters.samplerate
        filter_fc = self.cfg.settings.filter_fc

        if filt_type == "sos":
            # Second order sections, with padding
            padding_type = self.cfg.settings.filter_padding_type
            padding_length = self.cfg.settings.filter_padding_length
            if padding_length == -1:
                # Automatic padding length
                padlen = int(10 * fs / (2 * np.pi * filter_fc))
            filt_arg = (filt,)  # convert sos to a tuple (sos, ) to unpack it later
            filt_kwargs = dict(padtype=padding_type, padlen=padlen)
            filt_edge_method = "pad"
        elif filt_type == "ba":
            # Not-recommended scipy.signal.filtfilt() with Gustafsson’s method
            filter_ir_length = self.cfg.settings.filter_ir_length
            if filter_ir_length == -1:
                # Automatic impulse response length
                irlen = int(20 * fs / (2 * np.pi * filter_fc))
            filt_arg = filt  # already a tuple (b, a)
            filt_kwargs = dict(irlen=irlen)
            filt_edge_method = "gust"
        else:
            raise ValueError(
                f"{filt_type} is not a valid filter type, it must be 'sos' or 'ba'"
            )

        # Run the demodulation
        return sp.demodulate(
            reftime,
            ref,
            sigtime,
            sig,
            cos_func,
            *filt_arg,
            extra_phase=extra_phase,
            refas=refas,
            refphis=refphis,
            filter_edge_method=filt_edge_method,
            filter_kws=filt_kwargs,
        )

    @staticmethod
    def _find_phase_minoop(in_phase: np.ndarray, out_phase: np.ndarray) -> float:
        """
        Find phase that minimizes the integral of `out_phase`, keeping `in_phase` > 0.

        Parameters
        ----------
        in_phase, out_phase : np.ndarray
            I and Q signal vectors.

        Returns
        -------
        phase : float
            Phase.
        """

        # Define function to minimize
        def rot_oop(th: np.ndarray) -> float:
            return np.abs(np.sin(-th) * in_phase + np.cos(-th) * out_phase).sum() / len(
                in_phase
            )

        res = optimize.minimize_scalar(
            rot_oop, method="bounded", bounds=(-np.pi, np.pi)
        )
        ang = res.x
        if (np.cos(-ang) * in_phase - np.sin(-ang) * out_phase).sum() < 0:
            ang += np.pi
        # Wrap to -pi, +pi
        ang = np.arctan2(np.sin(ang), np.cos(ang))

        return ang

    def _find_phase(
        self,
        in_phase: np.ndarray,
        out_phase: np.ndarray,
        method: Literal["at peak field", "minimum oop"],
        idx_max: int = 0,
    ) -> float:
        """
        Find the phase shift that maximises signal in I.

        Parameters
        ----------
        in_phase, out_phase : np.ndarray
            I and Q signal vectors.
        method : {"at peak field", "minimum oop"}
            Method to find the phase. "at peak field" maximises the in-phase signal at
            peak field. "minimum oop" minimizes the integral of the out of phase signal
            while keeping a positive in-phase signal.
        idx_max : int, optional
            Index at maximum field. Required for "at peak field" method.
        """
        match method:
            case "at peak field":
                return np.arctan2(out_phase[idx_max], in_phase[idx_max])
            case "minimum oop":
                return self._find_phase_minoop(in_phase, out_phase)
            case _:
                logging.error(
                    f"Unkown method to find phase : {method}. Choose 'at peak field' or"
                    " 'minimum oop'"
                )

    def _check_data_loaded(self) -> bool:
        """Check whether the data was loaded."""
        return self.get_data_raw(self.signal_name, checkonly=True)

    def _check_demodulated(self) -> bool:
        """Check whether the signal has been demodulated."""
        return self.get_data_serie("in_phase", checkonly=True) and self.get_data_serie(
            "out_phase", checkonly=True
        )

    def _check_resistance_computed(self) -> bool:
        """Check whether the resistance was computed."""
        return self.get_data_serie(
            "resistance_in_phase", checkonly=True
        ) and self.get_data_serie("resistance_out_phase", checkonly=True)

    def _check_f0_computed(self) -> bool:
        """Check whether center frequency was found."""
        if self.signal_name in self.metadata["f0"]:
            if self.metadata["f0"][self.signal_name] > 0:
                return True
        return False



"""Signal processing module."""

import warnings
from collections.abc import Callable, Iterable
from typing import Any, Literal

import numpy as np
from scipy import integrate, ndimage, optimize, signal


def _sanitize_dims_for_demodulation(
    reftime: np.ndarray | None,
    refsig: np.ndarray | None,
    sigtime: np.ndarray,
    sig: np.ndarray,
) -> tuple[np.ndarray | None, np.ndarray | None, np.ndarray, np.ndarray]:
    # Reference time vector should be 1D
    if reftime is None:
        pass
    elif reftime.ndim != 1:
        raise ValueError(
            "Reference time vector has too much dimensions, it should be 1D."
        )
    # Others should be 2D even with only one serie for vectorization
    if refsig is None:
        refsig_na = None
    elif refsig.ndim == 1:
        refsig_na = refsig[..., np.newaxis]
    elif refsig.ndim == 2:
        refsig_na = refsig
    else:
        raise ValueError(
            "Reference signal has too much dimensions, it should be 1 or 2D."
        )
    if sigtime.ndim == 1:
        sigtime_na = sigtime[..., np.newaxis]
    elif (sigtime.ndim == 2) & (sigtime.shape[1] == 1):
        sigtime_na = sigtime
    else:
        raise ValueError(
            "Signal time vector has too much dimensions, it should be 1D "
            "or have its second dimension shape equal to 1."
        )
    if sig.ndim == 1:
        sig_na = sig[..., np.newaxis]
    elif sig.ndim == 2:
        sig_na = sig
    else:
        raise ValueError("Signal has too much dimensions, it should be 1 or 2D.")

    return reftime, refsig_na, sigtime_na, sig_na


def integrate_pickup(
    pickup_time: np.ndarray,
    pickup_signal: np.ndarray,
    pickup_surface: float,
    method: str = "trapz",
) -> np.ndarray:
    """
    Integrate pickup coil voltage to get the magnetic field.

    Uses Faraday-Lenz law : e = -SdB/dt. There could be multiple implementation to
    integrate, but only the cumulative trapezoid is implemented. The field is forced to
    be positive and starts at 0T.

    Parameters
    ----------
    pickup_time : np.ndarray
        1D vector with time points.
    pickup_signal : np.ndarray
        1D vector with pickup coil signal.
    pickup_surface : float
        Pickup coil surface in m².
    method : str, optional
        Method for the integration, by default "trapz".

    Returns
    -------
    magfield : np.ndarray
        Magnetic field in T.

    Raises
    ------
    ValueError
        Only 'trapz' is supported as integration `method`.

    """
    # Remove overall mean (offset) so that B is 0 when there's no magnetic field
    pu_signal = pickup_signal - np.mean(pickup_signal)

    # Integrate to get the field
    match method:
        case "trapz":
            magfield = (
                integrate.cumulative_trapezoid(pu_signal, pickup_time, initial=0)
                / pickup_surface
            )
        case _:
            err_msg = (
                "Only cumulative trapezoid integration method ('trapz') is implemented."
            )
            raise NotImplementedError(err_msg)

    # Fix inverted field
    if np.mean(magfield) < 0:
        magfield = -magfield

    return magfield


def rolling_average(
    a: np.ndarray, wlen: int, subsample: bool = False, axis=0
) -> np.ndarray:
    """
    Apply a rolling window average.

    The moving window is  applied on columns (with `axis=0`) or lines (with `axis=1`)
    of the input 2D array `a`. Optionally subsample the output.

    Parameters
    ----------
    a : np.ndarray
        2D array with series to smooth.
    wlen : int
        Time window size (expressed as indices).
    subsample : bool, optional
        If True, the series are sub-sampled `wlen - 1` times.
    axis : int, optional
        Direction to consider for time series, eg. 0 if time series to smooth are on the
        columns, 1 on lines. Default is 0.

    Returns
    -------
    smoothed_a : np.ndarray
        Array with smoothed time series.
    """
    # rolling average filter
    b = ndimage.uniform_filter1d(a, size=wlen, axis=axis, mode="reflect", origin=0)

    # subsampling, handling axes
    if subsample:
        b = subsample_array(b, wlen - 1, axis=axis)

    return b


def _subsample_array_direct(
    a: np.ndarray, step: int, start: int | None = None, axis: int = 0
) -> np.ndarray:
    """Subsampling with basic indexing."""
    if start is None:
        start = step
    if axis == 0:
        if a.ndim < 2:
            a = a[start::step].copy()
        else:
            a = a[start::step, :].copy()
    elif axis == 1:
        a = a[:, start::step].copy()
    else:
        raise NotImplementedError("Subsampling is not supported for axis > 1.")

    return a


def _subsample_array_average(
    a: np.ndarray, step: int, start: int | None = None, axis: int = 0
) -> np.ndarray:
    """Subsampling with average."""
    if a.ndim not in (1, 2):
        raise ValueError("a must be 1D or 2D.")

    if a.ndim == 1:
        # 1D
        pad_size = (step - (len(a) % step)) % step
        a_padded = np.pad(a, (0, pad_size), mode="constant", constant_values=np.nan)
        return np.nanmean(a_padded.reshape(-1, step), axis=1)

    # 2D
    pad_size = (step - (a.shape[axis] % step)) % step
    if axis == 0:
        a_padded = np.pad(
            a.astype(float),
            ((0, pad_size), (0, 0)),
            mode="constant",
            constant_values=np.nan,
        )
        return np.nanmean(a_padded.reshape(-1, step, a.shape[1]), axis=1)
    elif axis == 1:
        a_padded = np.pad(
            a.astype(float),
            ((0, 0), (0, pad_size)),
            mode="constant",
            constant_values=np.nan,
        )
        return np.nanmean(a_padded.reshape(a.shape[0], -1, step), axis=2)
    else:
        raise NotImplementedError("Subsampling is not supported for axis > 1.")


def subsample_array(
    a: np.ndarray,
    step: int,
    average: bool = False,
    start: int | None = None,
    axis: int = 0,
) -> np.ndarray:
    """
    Subsample an array in the given `axis` every `step` point.

    Optionnaly, take the mean of each slice instead of basic indexing.

    Parameters
    ----------
    a : np.ndarray
        Input array.
    step : int
        step::step points will be kept.
    average : bool, optional
            Use the mean of each slice. Default is False.
    start : int or None, optional
        Used only if `average` is False. Start at this index. If None, first index is
        `step` (default).
    axis : int < 2, optional
        Axis along which to operate, only 0 and 1 are supported. Default is 0 (along
        columns).

    Returns
    -------
    np.ndarray
        Subsampled array.

    Raises
    ------
    NotImplementedError
        If axis > 1.
    """
    if not average:
        return _subsample_array_direct(a, step, start=start, axis=axis)
    else:
        return _subsample_array_average(a, step, axis=axis)

    return a


def fit_signal(
    func: Callable[[np.ndarray, float | np.ndarray, float | np.ndarray], np.ndarray],
    x: np.ndarray,
    y: np.ndarray,
    **kwargs,
) -> tuple[np.ndarray, np.ndarray] | tuple[float, float]:
    """
    Fit `c0` and `c1` in `y = func(x, c0, c1)`.

    It performs an independant fit for each columns of `y`.

    Parameters
    ----------
    func : Callable
        Function taking a vector and 2 constants as arguments and returns a vector the
        same size as input.
    x : np.ndarray
        Time vector of `y`.
    y : np.ndarray
        Array with time series on columns.
    **kwargs : passed to `scipy.optimize.curve_fit()`.
    """
    if y.ndim == 1:
        popt, _ = optimize.curve_fit(func, x, y, **kwargs)
        c0 = popt[0]
        c1 = popt[1]
    elif y.ndim == 2:
        c0 = np.empty((y.shape[1],), dtype=float)
        c1 = np.empty((y.shape[1],), dtype=float)
        for frame in range(y.shape[1]):
            popt, _ = optimize.curve_fit(func, x, y[:, frame], **kwargs)
            c0[frame] = popt[0]  # reference amplitude
            c1[frame] = popt[1]  # reference phase

    return c0, c1


def demodulate_chunks(
    reftime: np.ndarray | None,
    refsig: np.ndarray | None,
    sigtime: np.ndarray,
    sig: np.ndarray,
    f0: float,
    filter_order: int,
    filter_fc: float,
    decimate_factor: int = 0,
    refas: float | np.ndarray | None = None,
    refphis: float | np.ndarray | None = None,
    chunksize: int = 0,
    bar: Callable[[Iterable], Iterable] | None = None,
    progress_emitter: Any = None,
) -> tuple[np.ndarray, np.ndarray]:
    """
    Demodulation with IQ real signals.

    Reference signal is fitted to get its amplitude and phase over time. Those are used
    to generate a continuous reference signal, used to demodulate `signal`. The I and Q
    components are returned.

    The signal is frequency shifted with the continuous reference and a low-pass filter
    is applied.

    The process is done in chunks to not overload the memory, calling the `demodulate()`
    function on each chunk.

    Optionally, decimation can be used to reduce the data before filtering.

    Parameters
    ----------
    reftime : np.ndarray
        Time vector of the reference signal.
    refsig : np.ndarray
        2D array of the reference signal, cropped around where the fit will be done,
        time series should be on columns.
    sigtime : np.ndarray
        Time vector of the whole signal.
    sig : np.ndarray
        2D array representing the signal. Time series should be on columns.
    f0 : float
        Center frequency (experimental RF frequency).
    filter_order : int
        Order of the low-pass filter.
    filter_fc : float
        Cut-off frequency of the low-pass filter.
    decimate_factor : int, optional
        Downsampling factor, set to 0 or 1 to disable (default). An anti-aliasing filter
        is applied before.
    chunksize : int, optional
        Size of chunks, set to -1 to disable or 0 to adapt chunk size so that there are
        100 chunks (default).
    bar : Callable or None, optional
        Decorator for the main loop. Should be a callable that takes an iterable as
        argument to display a progress bar. Tested for `rich.progress.track()` and
        `tqdm.tqdm()`). Default is None (no progress bar).
    progress_emitter : Any or None, optional
        An object with an `emit()` method, such as a pyqtSignal. The loop index (e.g.
        frames index) is emitted at each iteration of the main loop. Default is None.

    Returns
    -------
    I, Q : np.ndarray
        I-Q-Demodulated signal.
    """
    # Dimensions checks
    reftime, refsig, sigtime, sig = _sanitize_dims_for_demodulation(
        reftime, refsig, sigtime, sig
    )

    npoints, nframes = sig.shape
    if decimate_factor > 1:
        new_npoints = int(npoints / decimate_factor)
        new_shape = (new_npoints, nframes)
    else:
        new_shape = (npoints, nframes)

    # Find reference phase
    def cos_func(
        x: np.ndarray, A: float | np.ndarray, phi: float | np.ndarray
    ) -> np.ndarray:
        return A * np.cos(2 * np.pi * f0 * x + phi)

    # Initialize output arrays
    X = np.empty(new_shape, float)
    Y = np.empty(new_shape, float)

    # Design low-pass filter
    fs = 1 / np.mean(np.diff(sigtime, axis=0))
    if decimate_factor > 1:
        fs = fs / decimate_factor
    sos = signal.butter(filter_order, filter_fc, btype="lowpass", fs=fs, output="sos")

    # Determine chunk size
    chunks, cs = get_chunks(nframes, chunksize)

    if bar is not None:
        chunks = bar(chunks)

    # Processing in chunks
    for idx in chunks:
        # Select data
        s = slice(idx, idx + cs + 1)
        refsig_chunk = refsig[:, s] if refsig is not None else None
        sig_chunk = sig[:, s]

        X[:, s], Y[:, s] = demodulate(
            reftime,
            refsig_chunk,
            sigtime,
            sig_chunk,
            cos_func,
            sos,
            decimate_factor=decimate_factor,
        )

        if progress_emitter is not None:
            progress_emitter.emit(idx)

    return X, Y


def demodulate(
    reftime: np.ndarray | None,
    refsig: np.ndarray | None,
    sigtime: np.ndarray,
    sig: np.ndarray,
    func: Callable[[np.ndarray, float | np.ndarray, float | np.ndarray], np.ndarray],
    /,
    *args,
    extra_phase: float = 0,
    decimate_factor: int = 0,
    refas: float | np.ndarray | None = None,
    refphis: float | np.ndarray | None = None,
    filter_edge_method: Literal["pad", "gust"] = "pad",
    filter_kws: dict[str, Any] = {},
) -> tuple[np.ndarray, np.ndarray]:
    """
    Classical demodulation with IQ real signals.

    Reference signal is fitted to get amplitude and phase. Those are used to generate
    a continuous reference signal, used to demodulate `signal`. The I and Q components
    are returned.

    The signal is frequency shifted with the continuous reference and a low-pass filter
    is applied.

    The low-pass filter uses `scipy.signal.sosfiltfilt()` (cascaded second-order
    sections) if `filter_edge_method` is "pad". If `filter_edge_method` is "gust", it
    uses the Gustafsson’s method which is only implemented in `scipy.signal.filtfilt()`.
    If `filter_edge_method` is "pad", `sos` must be provided.
    If `filter_edge_method` is "gust", `a` and `b` must be provided.
    The filter is then configured with `filter_kws`.

    Optionally, decimation can be used to reduce the data before filtering. In that
    case, the filter must take into account the reduced sampling frequency.

    One can specify the reference amplitude and phases with the `refas` and `refphis`
    keyword arguments, in that case, the fit is not performed.

    Parameters
    ----------
    reftime : np.ndarray or None
        Time vector of the reference signal. Can be None if `refas` is specified.
    refsig : np.ndarray or None
        2D array of the reference signal with time series should be on columns. Can be
        None if `refphis` is specified.
    sigtime : np.ndarray
        Time vector of the whole signal.
    sig : np.ndarray
        2D array representing the signal with time series on columns.
    func : Callable
        Function taking a vector and 2 constants as arguments and returns a vector the
        same size as input. It is used to fit the reference so it should be a cosine
        function where the amplitude and the phase are fitted -- center frequency f0
        should be within the function.
    *args : tuple of np.ndarray
        Variable positional arguments:
        - If one array is provided, it is assigned to `sos`, and `a` and `b` are set to
        None. `sos` is the second-order sections representation of the IIR filter, as
        returned by `scipy.signal` filters with "sos" output. Used when
        `filter_edge_method` is "pad".
        - If two arrays are provided, they are assigned to `b` and `a`, and `sos` is set
        to None. `b` and `a` are the numerator and denominator coefficient vectors of
        the filter. Used when `filter_edge_method` is "gust".
    extra_phase : float, optional
        Add an additional phase offset (in rad) to the fitted reference phase. Default
        is 0 rad.
    decimate_factor : int, optional
        Downsampling factor, set to 0 to disable (default). An anti-aliasing filter is
        applied before.
    refas, refphis : np.ndarray, float or None, optional
        If specified, represent the amplitude and the phase of the cosine function. The
        fit is not performed to find the phase and the demodulation uses those instead.
    filter_edge_method : {"pad", "gust"}, optional
        Specify how to handle edges while applying the filter. "pad" uses padding, so
        second-order sections can be used (recommended, default). "gust" uses
        Gustafsson’s method, so `scipy.signal.filtfilt()` is used instead.
    filter_kws : dict[str, Any], optional
        Extra arguments for `scipy.signal.sosfiltfilt()` or `scipy.signal.filtfilt()`.

    Returns
    -------
    I, Q : np.ndarray
        I-Q-Demodulated signal.

    Examples
    --------
    With second order sections and padding
    ```
    sos = scipy.signal.butter(order, wn, btype="lowpass", output="sos")
    func = lambda x, a, phi: a * np.cos(2 * np.pi * f0 * x + phi)
    in_phase, out_phase = sp.demodulate(time, ref, time, sig, func, sos, extra_phase=0.01)
    ```
    With Gustafsson’s method :
    ```
    b, a = scipy.signal.butter(order, wn, btype="lowpass", output="ba")
    func = lambda x, a, phi: a * np.cos(2 * np.pi * f0 * x + phi)
    in_phase, out_phase = sp.demodulate(time, ref, time, sig, func, b, a, extra_phase=0.01)
    ```
    """
    # Handle variable positional arguments
    if len(args) == 1:
        sos, b, a = args[0], None, None
    elif len(args) == 2:
        sos, b, a = None, args[0], args[1]
    elif len(args) == 3:
        sos, b, a = args[0], args[1], args[2]
        if filter_edge_method == "gust":
            warnings.warn(
                "sos was provided but won't be used because filter_edge_method "
                "is 'gust'"
            )
        if filter_edge_method == "pad":
            warnings.warn(
                "b and a were provided but won't be used because filter_edge_method "
                "is 'pad'"
            )
    else:
        raise ValueError(
            "Too many arrays were provided for the filter coefficients. "
            "Use one array for sos or two arrays for b, a"
        )

    # Dimensions checks
    reftime, refsig, sigtime, sig = _sanitize_dims_for_demodulation(
        reftime, refsig, sigtime, sig
    )
    # Input checks and fit if ref amplitude and phase not provided
    if (refas is None) or (refphis is None):
        if (reftime is None) or (refsig is None):
            raise ValueError(
                "If reftime or refsig are None, refas and refphis must be specified"
            )
        else:
            refas, refphis = fit_signal(func, reftime, refsig)

    # Multiply signal by reference (frequency-shifting)
    x = func(sigtime, refas, refphis + extra_phase) * sig
    y = func(sigtime, refas, refphis + np.pi / 2 + extra_phase) * sig

    # Decimation (subsampling with anti-aliasing)
    if decimate_factor > 1:
        x = signal.decimate(x, decimate_factor, axis=0)
        y = signal.decimate(y, decimate_factor, axis=0)
    else:
        x = x
        y = y

    # Low-pass filter
    if filter_edge_method == "pad":
        if sos is None:
            raise ValueError("sos can't be None is filter_edge_method is 'pad'")
        x = signal.sosfiltfilt(sos, x, axis=0, **filter_kws)
        y = signal.sosfiltfilt(sos, y, axis=0, **filter_kws)
    elif filter_edge_method == "gust":
        if (a is None) or (b is None):
            raise ValueError("a and b can't be None if filter_edge_method is 'gust'")
        x = signal.filtfilt(b, a, x, axis=0, method="gust", **filter_kws)
        y = signal.filtfilt(b, a, y, axis=0, method="gust", **filter_kws)
    else:
        raise ValueError(
            f"{filter_edge_method} is not a valid method, choose 'pad' or 'gust'"
        )

    return x, y


def find_signal(
    sig: np.ndarray, std_factor: float, before: float = 0, after: float = 0
) -> tuple[int, int]:
    """
    Detect onset and offset of signal.

    Signal is defined as when the values are above/below the mean +/- `std_factor` times
    the standard deviation of the whole trace. Onset (offset) are the first (last, resp.)
    indices where this condition is met.
    `sig` can be 2D, in which case time series should be on column. The minimum index is
    selected for the onset and the last for offset.

    Parameters
    ----------
    sig : np.ndarray
        Array with time series on columns.
    std_factor : float
        Multipler of signal standard deviation.
    before : float, optional
        Fraction to take before the actual detected onset, by default 0.
    after : float, optional
        Fraction to take after the actual detected onset, by default 0.

    Returns
    -------
    start, stop : int
        Indices of onset and offset.
    """
    npoints = sig.shape[0]

    # Convert to float instead of doing it twice (mean and std)
    findsig = sig.astype(float)

    # Get threshold
    thresh = np.abs(findsig.mean(axis=0)) + std_factor * findsig.std(axis=0)

    # Binary mask
    mask = np.abs(findsig) > thresh

    # Find first non-zero value
    first = mask.argmax(axis=0).min()
    last = npoints - np.flip(mask, axis=0).argmax(axis=0).min()

    # Take a bit more or less
    delta = last - first
    first -= before * delta  # x% before onset
    last += after * delta  # x% after offset
    # Check boundaries
    first = max(0, int(first))
    last = min(int(last), npoints)
    # Check they didn't reverse, otherwise swap the two
    if last < first:
        ffirst = last
        last = first
        first = ffirst

    return first, last


def find_f0(sig: np.ndarray, fs: float) -> np.ndarray:
    """
    Find center frequency in signal `sig` sampled at `fs`.

    `sig` can be 2D, in which case the time series should be on columns, and the mean
    frequency found in each time series is returned.

    It uses the gaussian spectrum interpolation method described in Gasior & Gonzalez
    (2004) (1).

    (1) M. Gasior and J.L. Gonzalez, Improving FFT Frequency Measurement Resolution by
    Parabolic and Gaussian Spectrum Interpolation, CERN, 2004.

    Parameters
    ----------
    sig : np.ndarray
        S
    fs : float
        Sampling rate.

    Returns
    -------
    f0 : float
        Center frequency in same units as `fs`, one value per serie (columns of `sig`).
    """
    serie = sig.copy()
    npoints = serie.shape[0]

    # Check dimensions for compatibility with multiple time series
    if serie.ndim == 1:
        serie = serie[..., np.newaxis]

    # Indices vector
    time_ind = np.arange(npoints)[..., np.newaxis]

    # Gaussian coef.
    t0 = (time_ind[0, :] + time_ind[-1, :]) * 0.5
    c = 0.5 * (8 / (npoints - 1)) ** 2  # gaussian interp. with r=8
    gaussian = np.exp(-((time_ind - t0) ** 2) * c)

    # RFFT
    Sn = np.abs(np.fft.rfft(serie * gaussian, axis=0))
    f = np.fft.rfftfreq(npoints, d=1 / fs)

    # Frequency at FFT peak
    amax = Sn.argmax(axis=0)
    if np.all(0 < amax) & np.all(amax < f.shape[0]):
        indexer = np.arange(len(amax))
        dm = (
            fs
            / npoints
            * np.log(Sn[amax + 1, indexer] / Sn[amax - 1, indexer])
            / (
                2
                * np.log(
                    Sn[amax, indexer] ** 2
                    / (Sn[amax + 1, indexer] * Sn[amax - 1, indexer])
                )
            )
        )
        f0 = f[amax] + dm
    else:
        # Too close to the boundaries of the FFT window
        warnings.warn(
            "Center frequency is at the boundary of the window, "
            "can't interpolate properly."
        )
        f0 = f[amax]

    return f0


def compute_amp_iq(in_phase: np.ndarray, out_phase: np.ndarray) -> np.ndarray:
    """
    Compute amplitude from I and Q.

    Parameters
    ----------
    in_phase, out_phase : np.ndarray
        Same sized arrays corresponding to I and Q respectively.

    Returns
    -------
    amplitude : np.ndarray
        Signal amplitude.
    """
    return np.sqrt(in_phase**2 + out_phase**2)


def compute_phase_iq(
    in_phase: np.ndarray,
    out_phase: np.ndarray,
    unwrap: bool = False,
    period: float = np.pi,
    axis: int = 1,
) -> np.ndarray:
    """
    Compute phase from I and Q.

    The result can be unwrapped, in which case an axis must be specified.

    Parameters
    ----------
    in_phase, out_phase : np.ndarray
        Same sized arrays corresponding to I and Q respectively.
    unwrap : bool, optional
        Unwrap resulting phase. Default is False.
    period : float, optional
        Period considered for unwrapping, used only if `unwrap` is True. Default is pi.
    axis : int
        If `unwrap` is True, defines in which axis are time series, default is 1 (time
        series on columns).

    Returns
    -------
    phase : np.ndarray
        Signal phase.
    """
    res = np.arctan2(out_phase, in_phase)

    if unwrap:
        res = np.unwrap(res, period=period, axis=axis)

    return res


def rescale_a2b(
    a: np.ndarray, b: np.ndarray, allow_offset: bool = False, sub_mean: bool = False
) -> np.ndarray:
    """
    Rescale `a` so that it is in the same range as `b`.

    If `allow_offset` is False (default), the minimum will be set at 0 instead. If
    `sub_mean` is True, the mean is subtracted from the final array before returning.

    Parameters
    ----------
    a : np.ndarray
        Array to rescale.
    b : np.ndarray
        Array to get range from.
    allow_offset : bool, optionnal
        If False (default), the minimum value of the returned array is 0, otherwise, it
        is the minimum of `b`.
    sub_mean : bool, optionnal
        If True, the mean is subtracted from the rescaled array. Default is False.

    Returns
    -------
    rescaled_a : np.ndarray
        `a`, rescaled.

    """
    acopy = a.copy().astype(float)
    amin, amax = acopy.min(), acopy.max()
    bmin, bmax = b.min().astype(float), b.max().astype(float)
    rescaled_a = ((acopy - amin) / (amax - amin)) * (bmax - bmin)
    if allow_offset:
        rescaled_a += bmin
    if sub_mean:
        rescaled_a -= rescaled_a.mean()

    return rescaled_a


def get_chunks(nitems, chunk_size) -> tuple[range, int]:
    """
    Get a range generator.

    The generator goes from 0 to `nitems` with steps of `chunk_size`.
    If `chunk_size` is -1, it goes from 0 to `nitems` in one step (i.e. no chunking).
    If `chunk_size` is 0, it adjusts the step size to get 100 chunks.

    Parameters
    ----------
    nitems : int
        Total number of items.
    chunk_size : int
        Step size. -1 and 0 are treated as special values (see description above).

    Returns
    -------
    range : range
        Range generator.
    step_size : int
        The step size (chunk size) that was determined.
    """
    # Determine chunk size
    if chunk_size == -1:
        # no chunking
        cs = nitems
    elif chunk_size == 0:
        # get 100 chunks
        cs = nitems // 100
    else:
        cs = chunk_size

    return range(0, nitems, cs), cs


def compute_attenuation(
    amp: np.ndarray,
    amp0: float,
    echo_idx: int,
    length: float,
    mode: Literal["reflection", "transmission"] = "reflection",
    scale: Literal["linear", "log"] = "linear",
    corr: float = 1.0,
) -> np.ndarray:
    """
    Compute the normalized amplitude : attenuation in dB/m.

    The attenuation in dB/m is defined as :
        (-20 / (2nL)) * log10(A(B) / A(0))
    With n the echo index, L the sample length, A(B) the amplitude (that depends on the
    magnetic field) and A(0) the amplitude without magnetic field.

    In transmission `mode`, the wave propagates in the sample "one-way", in reflection
    the round trip is taken into account.

    Because in some setups there might be a logarithmic amplificator, `amp` might
    already be in logarithmic scale. In that case, the `log10` in the formula are
    removed, and the amplificator slope (`corr`) needs to be corrected for. To do so,
    set `scale` to "log".

    Parameters
    ----------
    amp : np.ndarray
        Amplitude time serie.
    amp0 : float
        Amplitude baseline.
    echo_idx : int
        Index of the analyzed echo (1-based, the first echo is 1).
    length : float
        Sample length in m.
    mode : {"reflection" , "transmission"}, optional
        Detection mode, by default "reflection".
    scale: {"linear", "scale"}, optional
        Scaling of the attenuation: linear scale or log scale
    corr : float, optional
        Denominator correction factor (logarithmic amplificator slope). Default is 1
        (no correction). Used only if scale is "log".

    Returns
    -------
    attenuation : np.ndarray
        Normalized amplitude in dB/m.

    """
    match mode:
        case "reflection":
            # reflection : 2nL
            constant = 2 * echo_idx

        case "transmission":
            # transmission : (2n - 1)L
            constant = 2 * echo_idx - 1

        case _:
            raise ValueError(
                "Expected 'reflection' or 'transmission' for detection mode, got"
                f" {mode}"
            )

    match scale:
        case "linear":
            return -20 * np.log10(amp / amp0) / (constant * length)
        case "log":
            return (amp0 - amp) / (corr * constant * length)
        case _:
            raise ValueError(
                f"Expected 'linear' or 'transmission' for 'scale', got '{scale}'"
            )


def compute_phase_shift(
    phi: np.ndarray,
    phi0: float,
    echo_idx: int,
    speed: float,
    rf_freq: float,
    length: float,
    mode: str = "reflection",
) -> np.ndarray:
    """
    Compute relative phase-shift.

    Parameters
    ----------
    phi : np.ndarray
        Pi-jump-corrected phase time serie.
    phi0 : float
        Phase baseline.
    echo_idx : int
        Index of analyzed echo (1-based).
    speed : float
        (Estimated) speed of sound in the sample.
    rf_freq : float
        Radiofrequency frequency used in the experiment in Hz.
    length : float
        Sample length in m.
    mode : {"reflection", "transmission"}, optional
        Detection mode, by default "reflection".

    Returns
    -------
    deltaphi : np.ndarray
        Relative phase-shift.

    """
    match mode:
        case "reflection":
            # reflection : 2nL
            constant = 2 * echo_idx
        case "transmission":
            # transmission (2n - 1)L
            constant = 2 * echo_idx - 1
        case _:
            raise ValueError(
                "Expected 'reflection' or 'transmission' for detection mode. "
                f"Got {mode}."
            )

    return (phi - phi0) * speed / (2 * np.pi * rf_freq * constant * length)


def _find_barycenters_fast(
    fxx: np.ndarray, sxx: np.ndarray, freq_window: float
) -> np.ndarray:
    max_f_idx = np.argmax(sxx, axis=0)
    bary_win_size = int(freq_window / np.mean(np.diff(fxx)))
    # sxx is nfreqs x ntimes
    # Create indices for all freq-windows at each time so dimensions :
    # winsize x ntimes
    # Indices are -30:30 centered on each max freq
    indices = (
        np.arange(-bary_win_size, bary_win_size + 1)[..., np.newaxis]
        + max_f_idx[np.newaxis, ...]
    )
    # Make sure indices are in range
    indices = np.clip(indices, 0, sxx.shape[0] - 1)
    # Extract windows for all columns. We need cols_indexer to select all columns of
    # the spectrogram and pair them with the frequency indices
    cols_indexer = np.arange(sxx.shape[1])
    sxx_windows = sxx[indices, cols_indexer]
    fxx_windows = fxx[indices]
    # Compute barycenters
    sum_weight_f = np.sum(fxx_windows * sxx_windows, axis=0)
    sum_weight = np.sum(sxx_windows, axis=0)

    return sum_weight_f / sum_weight


def _find_barycenters_slow(
    fxx: np.ndarray, sxx: np.ndarray, freq_window: float
) -> np.ndarray:
    max_f_idx = np.argmax(sxx, axis=0)
    bary_win_size = int(freq_window / np.mean(np.diff(fxx)))
    barycenters = np.zeros(max_f_idx.size)
    for idx in range(max_f_idx.size):
        indexer = slice(max_f_idx[idx] - bary_win_size, max_f_idx[idx] + bary_win_size)
        sum_weight_f = np.sum(fxx[indexer] * sxx[indexer, idx])
        sum_weight = np.sum(sxx[indexer, idx])
        barycenters[idx] = sum_weight_f / sum_weight

    return barycenters


def find_barycenters(
    fxx: np.ndarray, sxx: np.ndarray, freq_window: float, fast: bool = True
) -> np.ndarray:
    """
    Compute the barycenters of the maximum frequencies over time in a spectrogram.

    It uses the spectrogram from `scipy.signal.spectrogram()`.

    Parameters
    ----------
    fxx : np.ndarray
        Array with the frequencies of the spectrogram, as returned by
        `scipy.signal.spectrogram()`.
    sxx : np.ndarray
        Spectrogram as returned by `scipy.signal.spectrogram()`.
    freq_window : float
        Frequency-window size around the frequency maxima, in the same units as `fxx`.
    fast : bool, optional
        Whether to use the vectorized version of the algorithm, default is True.
    """
    if fast:
        barycenters = _find_barycenters_fast(fxx, sxx, freq_window)
    else:
        barycenters = _find_barycenters_slow(fxx, sxx, freq_window)

    return barycenters


def find_nearest_index(array: np.ndarray, value: float) -> int:
    """
    Find the index of the nearest value in `array` to the input `value`.

    Parameters
    ----------
    array : np.ndarray
        Array from which the index is extracted from.
    value : float
        Value to find in `array`.
    """
    return (np.abs(array - value)).argmin()


def get_up_down_indices(x: np.ndarray) -> tuple[slice, slice]:
    """
    Get indices in `y` for increasing and decreasing `x`.

    Returns 2 slices, the first indexes `y` in increasing `x`, the second in decreasing
    `x`. `x` should be a vector that increases, reach a maximum, then decreases.

    Parameters
    ----------
    x : np.ndarray
        Should be 1D.

    Returns
    -------
    slice_inc, slice_dec : slice
    """
    idx_at_xmax = np.argmax(x)
    return (slice(0, idx_at_xmax), slice(-1, idx_at_xmax - 1, -1))


def split_up_down(
    x: np.ndarray, y: np.ndarray
) -> tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    """
    Split `y` in two arrays based on if `x` is increasing or decreasing.

    `x` should be a vector that increases, reaches a maximum and then decreases (e.g. a
    magnetic field).
    One array corresponds to increasing `x`, the othe corresponds to decreasing `x`. The
    arrays are sorted and the `x` arrays are also returned.

    Parameters
    ----------
    x, y : np.ndarray
        1D vectors.

    Returns
    -------
    x_inc, y_inc : np.ndarray
        `x` and `y` when `x` increases.
    x_dec, y_dec : np.ndarray
        `x` and `y` when `x` decreases.
    """
    s_inc, s_dec = get_up_down_indices(x)
    return (x[s_inc], y[s_inc], x[s_dec], y[s_dec])


def fit_poly(
    field: np.ndarray,
    sig: np.ndarray,
    poly_deg: int,
    boundary1: float | None = None,
    boundary2: float | None = None,
) -> tuple[np.ndarray, np.ndarray]:
    """
    Polynomial fit of `sig(field)`, between boundaries.

    Returns the resulting polynome evaluated at `field` along with the detrended signal.

    Parameters
    ----------
    field : np.ndarray
        Magnetic field vector (x).
    sig : np.ndarray
        Signal vector (y).
    poly_deg : int
        Polynome degree.
    boundary1, boundary2 : float or None, optional
        Do the fit only in `field` comprised in `[poly_boundary1, poly_boundary2]`. If
        None (default), use all the data.

    Returns
    -------
    poly : np.ndarray
        Polynome evaluated at `field`.
    sig_detrend : np.ndarray
        `sig` subtracted
    """
    if boundary1 is None:
        idx_bmin = 0
    else:
        idx_bmin = find_nearest_index(field, boundary1)
    if boundary2 is None:
        idx_bmax = len(field)
    else:
        idx_bmax = find_nearest_index(field, boundary2)

    # Collect data
    x = field[idx_bmin:idx_bmax]
    y = sig[idx_bmin:idx_bmax]
    # Check data
    if x.size < 2:
        raise ValueError(
            f"x has only {x.size} elements with bounds {idx_bmin, idx_bmax}"
        )
    if y.size < 2:
        raise ValueError(
            f"y has only {y.size} elements with bounds {idx_bmin, idx_bmax}"
        )

    # Fit
    fit = np.polynomial.Polynomial.fit(x, y, poly_deg)
    res = fit(field)
    return res, sig - res


def interpolate_inverse(
    x: np.ndarray, y: np.ndarray, boundary1: float, boundary2: float, npoints: int
) -> tuple[np.ndarray, np.ndarray]:
    """
    Oversample `y` in `1/x` in the given range.

    Parameters
    ----------
    x : np.ndarray
        Points at which `y` is evaluated.
    y : np.ndarray
        Signal to oversample.
    boundary1, boundary2 : float
        Define the range to select and oversample, before inversion.
    npoints : int
        Number of points of the resulting vectors.

    Returns
    -------
    x_inverse_oversample : np.ndarray
        `1/x` oversampled in `[1/boundary2, 1/boundary1]`.
    y_oversample : np.ndarray
        `y` oversampled in `[1/boundary2, 1/boundary1]`.
    """
    # Create the oversampled time vector
    x_inverse_oversample = np.linspace(
        1 / boundary2, 1.0 / boundary1, npoints, endpoint=False
    )
    # Interpolate
    y_oversample = np.interp(1 / x_inverse_oversample, x, y)

    return x_inverse_oversample, y_oversample


def fourier_transform(
    a: np.ndarray, d: float, pad_mult: float = 1
) -> tuple[np.ndarray, np.ndarray]:
    """
    Fourier transform for 1D real signal, with extra padding.

    Parameters
    ----------
    a : np.ndarray
        Input array.
    d : float
        Sample spacing.
    pad_mult : float, optional
        Multiplier of signal size used for padding. Default is 1.

    Returns
    -------
    freq : np.ndarray
        Frequencies.
    X : np.ndarray
        Magnitude of the fourier transform.
    """
    n = int(a.size * pad_mult)
    X = np.fft.rfft(np.hamming(a.size) * a, n=n)[:-1]
    freq = np.fft.rfftfreq(n, d=d)[:-1]
    X = np.abs(X) * 2 / n
    return freq, X


def collate_arrays(to_save: list[np.ndarray]):
    """
    Store 1D vectors as columns in a 2D arrays.

    Dimension mismatches are filled with trailing NaNs.

    Parameters
    ----------
    to_save : list[np.ndarray]
        List of arrays to save in the same file.

    Returns
    -------
    a : np.ndarray
        The array with all the vectors, with NaNs to fill shorter vectors.
    """
    # Build the array
    ncolumns = len(to_save)
    nrows = max([a.size for a in to_save])
    a_out = np.full((nrows, ncolumns), np.nan)

    # Fill the array
    for idx_col, a in enumerate(to_save):
        a_out[: a.size, idx_col] = a

    return a_out



"""The widget that hosts the grid with the main action buttons for pyuson."""

from PyQt6 import QtWidgets
from PyQt6.QtCore import Qt, pyqtSignal


class ButtonsWidget(QtWidgets.QWidget):
    """
    The main action buttons.

    pyqtSignals
    -------
    sig_load : emits when the "Load data" button is clicked.
    sig_show : emits when the "Show frames" button is clicked.
    sig_rolling : emits when the "Rolling average" button is clicked.
    sig_refresh : emits when the "Refresh" button is clicked.
    sig_save_csv : emits when the "Export as CSV" button is clicked.
    sig_save_nexus : emits when the "Save as NeXus" button is clicked.
    """

    sig_load = pyqtSignal()
    sig_show = pyqtSignal()
    sig_rolling = pyqtSignal()
    sig_refresh = pyqtSignal()
    sig_save_csv = pyqtSignal()
    sig_save_nexus = pyqtSignal()

    def __init__(self) -> None:
        super().__init__()
        """Initialize buttons."""
        # Create grid
        grid = QtWidgets.QGridLayout()

        # Create buttons
        # Load data
        self.button_load = QtWidgets.QPushButton("Load data", self)
        # Show some frames
        self.button_show = QtWidgets.QPushButton("Show frames", self)
        # Apply moving mean filter
        self.button_rolling = QtWidgets.QPushButton("Rolling average", self)
        # Trigger computation of attenuation and phase shift
        self.button_refresh = QtWidgets.QPushButton("Refresh", self)
        # Export as CSV
        self.button_save_csv = QtWidgets.QPushButton("Export as CSV", self)
        # Checkbox to convert to dB/cm
        self.checkbox_to_cm = QtWidgets.QCheckBox("Convert to dB/cm in CSV", self)
        self.checkbox_to_cm.setChecked(True)

        # Save as NeXus
        self.button_save_nexus = QtWidgets.QPushButton("Save as NeXus", self)

        # Disable all buttons
        self.disable_buttons()

        # Connect them
        self.connect_buttons()

        # Add buttons to layout
        grid.addWidget(self.button_load, 0, 0)
        grid.addWidget(self.button_show, 0, 1)
        grid.addWidget(self.button_rolling, 1, 0)
        grid.addWidget(self.button_refresh, 1, 1)
        grid.addWidget(self.button_save_csv, 2, 0)
        grid.addWidget(self.button_save_nexus, 2, 1)
        grid.addWidget(self.checkbox_to_cm, 3, 0, Qt.AlignmentFlag.AlignHCenter)

        # Set widget layout
        self.setLayout(grid)

    def connect_buttons(self) -> None:
        """Connect the buttons to their signals."""
        self.button_load.clicked.connect(self.sig_load.emit)
        self.button_show.clicked.connect(self.sig_show.emit)
        self.button_rolling.clicked.connect(self.sig_rolling.emit)
        self.button_refresh.clicked.connect(self.sig_refresh.emit)
        self.button_save_csv.clicked.connect(self.sig_save_csv.emit)
        self.button_save_nexus.clicked.connect(self.sig_save_nexus.emit)

    def disable_buttons(self) -> None:
        """Disable all buttons."""
        self.button_load.setEnabled(False)
        self.button_rolling.setEnabled(False)
        self.button_show.setEnabled(False)
        self.button_refresh.setEnabled(False)
        self.button_save_csv.setEnabled(False)
        self.button_save_nexus.setEnabled(False)

    def enable_buttons(self) -> None:
        """Enable all buttons (except Load data)."""
        self.button_load.setEnabled(False)
        self.button_rolling.setEnabled(True)
        self.button_show.setEnabled(True)
        self.button_refresh.setEnabled(True)
        self.button_save_csv.setEnabled(True)
        self.button_save_nexus.setEnabled(True)



"""Parameters for the Linx app."""

import numpy as np

from pymagnetos.core.gui.widgets import BaseParamContent


class ParamContent(BaseParamContent):
    PARAMS_TO_PARSE = []
    children_parameters = [
        dict(
            name="pickup_surface",
            type="float",
            limits=[0, np.inf],
            value=None,
            suffix="m²",
            siPrefix=False,
            readonly=True,
            title="Pickup surface",
        ),
        dict(
            name="pickup_samplerate",
            type="float",
            limits=[0, np.inf],
            value=500e3,
            step=1e3,
            suffix="Hz",
            siPrefix=True,
            readonly=True,
            title="Pickup sample rate",
        ),
        dict(
            name="current",
            type="float",
            limits=[0, np.inf],
            value=1,
            step=1,
            suffix="A",
            siPrefix=True,
            readonly=True,
            title="Pickup sample rate",
        ),
        dict(
            name="gain",
            type="float",
            limits=[0, np.inf],
            value=250e3,
            step=1e3,
            siPrefix=False,
            readonly=True,
            title="Ampli. gain",
        ),
        dict(
            name="gain_rms",
            type="bool",
            value=False,
            readonly=True,
            title="Ampli. gain RMS",
        ),
    ]

    children_settings = [
        dict(
            name="ref_name",
            type="list",
            limits=[],
            value="",
            title="Ref. name",
        ),
        dict(
            name="signal_name",
            type="list",
            limits=[],
            value="",
            title="Signal name",
        ),
        dict(
            name="time_offset",
            type="float",
            limits=[-np.inf, np.inf],
            value=0e-6,
            step=0.5e-6,
            suffix="s",
            siPrefix=True,
            title="Signal/Pickup time offset",
        ),
        dict(
            name="f0",
            type="float",
            limits=[0, np.inf],
            step=1e2,
            value=1e4,
            suffix="Hz",
            siPrefix=True,
            title="Demodulation rate"
        ),
        dict(
            name="phase_fit_ncycles",
            type="int",
            limits=[1, np.inf],
            step=1,
            value=10,
            title="Fit phase : number of periods",
        ),
        dict(
            name="filter_order",
            type="int",
            limits=[1, np.inf],
            step=1,
            value=2,
            title="Filter : order",
        ),
        dict(
            name="filter_fc",
            type="float",
            limits=[0, np.inf],
            step=1,
            value=1e3,
            suffix="Hz",
            siPrefix=True,
            title="Filter : cut-off frequency",
        ),
        dict(
            name="phase_offset",
            type="float",
            limits=[-np.inf, np.inf],
            step=1e-3,
            value=0,
            suffix="rad",
            siPrefix=False,
            title="Ref. phase offset"
        ),
        dict(
            name="auto_phase_method",
            type="list",
            value="at peak field",
            limits=["at peak field", "minimum oop"],
            title="Auto-phase method",
        ),
    ]



"""TDO Processor class to analyse TDO experiments."""

import logging
from collections.abc import Callable
from pathlib import Path
from typing import Literal, Self

import numpy as np
from scipy import signal

from ..core import BaseProcessor, sp
from ..log import configure_logger
from ._config import TDOConfig

MEAS_TO_CROP = (
    "pickup_time",
    "pickup",
    "magfield_time",
    "magfield",
    "signal_time",
)

TDO_SIGNAL_NAME = "tdo_signal"  # name of the TDO signal (the frequencies barycenters)
CSV_TDO_NCOLS = 3
CSV_RES_NCOLS = 11

logger = logging.getLogger(__name__)


class TDOProcessor(BaseProcessor):
    """A Processor class for TDO experiments."""

    def __init__(self, *args, **kwargs) -> None:
        configure_logger(logger, "pytdo.log")

        self._config_cls = TDOConfig

        super().__init__(*args, **kwargs)

        self.metadata = dict()

        self._meas_name = [*self.cfg.measurements.keys()][0]
        self._tdo_name = TDO_SIGNAL_NAME
        self._tdo_det_inc_name = self._tdo_name + "_detrend_inc"
        self._tdo_det_dec_name = self._tdo_name + "_detrend_dec"
        self._tdo_inv_inc_name = self._tdo_det_inc_name.replace("detrend", "inverse")
        self._tdo_inv_dec_name = self._tdo_det_dec_name.replace("detrend", "inverse")

        self.npoints_raw = 0

        self.is_cropped = {meas_name: False for meas_name in MEAS_TO_CROP}
        self.is_cropped[self._meas_name] = False

    def _reinit(self) -> None:
        """Clear data and initialize objects."""
        super()._reinit()

        self.npoints_raw = 0
        self.is_cropped = {meas_name: False for meas_name in MEAS_TO_CROP}
        self.is_cropped[self._meas_name] = False
        self.inds_b_inc = slice(-1)
        self.inds_b_dec = slice(-1)

    def load_pickup(self) -> Self:
        """
        Load the pickup binary file with the settings from the configuration.

        The experiment temperature and the sampling rate is also read from the file and
        added to the metadata.
        """
        # Get reader settings
        filename = Path(self.cfg.filenames["pickup"])
        precision = self.cfg.files["pickup"].precision
        endian = self.cfg.files["pickup"].endian
        order = self.cfg.files["pickup"].order
        npickups = self.cfg.parameters.pickup_number
        pickup_index = self.cfg.parameters.pickup_index

        logger.info(f"Loading pickup data from from {filename.name}...")
        if filename.is_file():
            # Load file
            pu = self._load_pickup(
                filename,
                precision=precision,
                endian=endian,
                order=order,
                nseries=npickups,
                index=pickup_index,
            )
            # Store metadata
            self.metadata["fs_pickup"] = float(pu[0])
            self.metadata["temperature"] = float(pu[1])

            # Store in the Data object
            nsamples = pu.shape[0] - 2
            self.set_data_raw(
                "pickup_time",
                np.linspace(
                    0, nsamples / self.metadata["fs_pickup"], nsamples, endpoint=False
                ),
            )
            self.set_data_raw("pickup", pu[2:])

            logger.info(
                f"Pickup loaded. Temperature : {self.metadata['temperature']:2.4f}K, "
                f"{nsamples} points sampled at {self.metadata['fs_pickup'] / 1e3}kHz"
            )
        else:
            # We'll simulate a fake pickup when we'll know the signal size
            logger.info(f"{filename.name} not found, skipping")
            self.set_data_raw("pickup_time", np.array([]))
            self.set_data_raw("pickup", np.array([]))
            self.metadata["fs_pickup"] = 1.0

        self.is_cropped["pickup_time"] = False
        self.is_cropped["pickup"] = False

        return self

    def compute_field(self, method: str = "trapz") -> Self:
        """Compute magnetic field."""
        super().compute_field(method=method)

        self.is_cropped["magfield"] = False
        self.is_cropped["magfield_time"] = False

        return self

    def load_data(self) -> Self:
        """Load raw data."""
        self.load_oscillo()
        return self

    def preprocess(self) -> Self:
        """Extract TDO signal from raw data."""
        self.extract_tdo()
        return self

    def align_field(self, on_pickup: bool = False, **kwargs) -> Self:
        """
        Align magnetic field on the `time_exp` vector.

        **kwargs are ignored.
        """
        if not self._check_field_computed():
            self.compute_field()
            self.crop_signal()

        if not self._check_barycenters_computed():
            self.find_barycenters()
            if not self._check_barycenters_computed():
                logger.error("TDO signal could not be extracted, check messages above")
                return self

        super().align_field(on_pickup=on_pickup, left=0)

        # Remove previously-computed values based on this TDO signal
        self._remove_data_changed()

        return self

    def run_analysis(self) -> Self:
        """Detrend the signal, sort signal in 1/B and compute FFT."""
        self.detrend_polyfit().oversample_inverse_field().fft_inverse_field()

        return self

    def export_tdo_csv(self, filename: str | Path | None = None, sep: str = "\t") -> Self:
        """
        Save intermediate results as a CSV file.

        The intermediate results correspond to the extracted TDO signal, before any
        post-processing (detrending, ...).
        If the file already exists, it is overwritten.
        """
        if filename is None:
            filename = self.get_csv_filename(suffix="-out")

        logger.info("Saving extracted TDO signal...")

        if not self._check_field_aligned():
            logger.warning(
                "The TDO signal and the magnetic fields are not aligned, aligning now"
            )
            self.align_field()

        to_save = ["time_exp", "magfield", self._tdo_name]
        a = sp.collate_arrays(self.collect_arrays(to_save))

        header = sep.join(["time", "field", "tdo"])

        np.savetxt(
            filename,
            a,
            delimiter=sep,
            header=header,
            comments="",
        )

        logger.info(f"Saved at {Path(filename).name}")

        return self

    def export_results_csv(
        self,
        filename_prefix: str | Path | None = None,
        sep: str = "\t",
    ) -> Self:
        """
        Save analysed results as CSV.

        It includes : time, aligned magnetic field, TDO signal, detrended TDO signal,
        1/B, TDO signal oversampled in 1/B, Fourier transform.
        """
        if filename_prefix is None:
            filename_prefix = Path(self.get_csv_filename(suffix="-results"))
        else:
            filename_prefix = Path(filename_prefix)

        logger.info("Saving analysed results...")

        # Increasing magnetic field
        filename_inc = filename_prefix.with_stem(filename_prefix.stem + "-up")
        to_save, header = self._get_results_up_down("inc")
        header_inc = sep.join(header)
        np.savetxt(
            filename_inc,
            to_save,
            delimiter=sep,
            header=header_inc,
            comments="",
        )

        # Decreasing magnetic field
        filename_dec = filename_prefix.with_stem(filename_prefix.stem + "-down")
        to_save, header = self._get_results_up_down("dec")
        header_dec = sep.join(header)
        np.savetxt(
            filename_dec,
            to_save,
            delimiter=sep,
            header=header_dec,
            comments="",
        )

        logger.info(f"Saved at {filename_inc.name} and {filename_dec.name}")

        return self

    def load_csv(self, filepath: Path | str, sep: str | None = None) -> Self:
        """Load data from a CSV file."""
        filepath = Path(filepath)
        if not filepath.is_file():
            logger.error("File does not exist")
            return self

        reader, delimiter = self._determine_file_format(filepath, sep=sep)

        if isinstance(reader, str) and reader == "unknown":
            logger.error(f"Could not load input file ({filepath})")
            return self

        logger.info(f"Reading data from {filepath.name}...")
        self._reinit()
        assert not isinstance(reader, str)  # for type checking
        status = reader(filepath, delimiter)
        if not status:
            logger.error(f"Failed to read {filepath.name}, check messages above")
            return self

        # Get the up / down indices
        self.get_up_down_indices()

        logger.info(f"Read data from {filepath.name}")

        return self

    def load_oscillo(self) -> Self:
        """
        Load TDO signal.

        Corresponding data is stored in `data_raw`.
        """
        # Get reader settings
        filename = Path(self.cfg.filenames["oscillo"])
        precision = self.cfg.files["oscillo"].precision
        endian = self.cfg.files["oscillo"].endian

        if not filename.is_file():
            logger.error(f"{filename.name} not found.")
            return self

        # Load data
        logger.info(f"Loading data from {filename.name}...")
        data = self.load_bin(filename, precision=precision, endian=endian).astype(float)

        # Store metadata
        self.metadata["fs_signal"] = float(data[0])
        temperature = float(data[1])
        self.npoints_raw = data.shape[0] - 2

        if "temperature" in self.metadata:
            # Already the temperature found in the pickup binary file
            other_temperature = self.metadata.pop("temperature")
            if temperature != other_temperature:
                logger.error(
                    "Temperatures found in the files do not match "
                    f"(got {temperature} and {other_temperature})"
                )

        # Keep the one from the TDO binary file
        self.metadata["temperature"] = temperature

        # Store data
        self.set_data_raw(self._meas_name, data[2:])
        self.get_signal_time()

        self.is_cropped[self._meas_name] = False

        logger.info(
            f"{self._meas_name} loaded. "
            f"Temperature : {temperature:2.4f}K, {self.npoints_raw} points sampled"
            f" at {self.metadata['fs_signal'] / 1e6}MHz"
        )

        return self

    def get_signal_time(self, fs: float | None = None) -> Self:
        """Get the experiment time vector corresponding to the raw data."""
        if fs is None:
            fs = self.metadata["fs_signal"]
        elif "fs_signal" not in self.metadata:
            self.load_oscillo()

        self.set_data_raw(
            "signal_time",
            np.linspace(
                0,
                self.npoints_raw / fs,
                self.npoints_raw,
                endpoint=False,
            ),
        )
        self.is_cropped["signal_time"] = False

        return self

    def crop_signal(self, force: bool = False) -> Self:
        """
        Crop signal up until `max_time` defined in the settings.

        The resulting signal is put in `data_raw`, the original data is kept with a
        trailing `_full`.
        """
        logger.info("Cropping data...")

        # pyqtSignal
        if "fs_signal" in self.metadata:
            idx_crop = int(self.metadata["fs_signal"] * self.cfg.settings.max_time)
            self._crop_signal(self._meas_name, 0, idx_crop, where="raw", force=force)
            self._crop_signal("signal_time", 0, idx_crop, where="raw", force=force)
        else:
            logger.warning("Data not loaded")

        # Pickup and magnetic field
        if "fs_pickup" in self.metadata:
            idx_crop = int(self.metadata["fs_pickup"] * self.cfg.settings.max_time)
            self._crop_signal("pickup_time", 0, idx_crop, where="raw", force=force)
            self._crop_signal("pickup", 0, idx_crop, where="raw", force=force)
            self._crop_signal(
                "magfield_time", 0, idx_crop, where="processed", force=force
            )
            self._crop_signal("magfield", 0, idx_crop, where="processed", force=force)
        else:
            logger.warning("Pickup not loaded")

        return self

    def compute_spectrogram(self) -> Self:
        """
        Perform a spectrogram on the raw signal.

        This a FFT performed on overlapping time windows of the signal. This results in
        a power spectral density (PSD) as a function of time and frequency.
        Parameters for the spectrogram are defined in the settings section of the
        configuration.

        The time and frequency vector and the 2D PSD are stored in `data_processed`.
        """
        # Check if data was loaded
        if not self._check_data_loaded():
            self.load_oscillo()
        # Check if data was time-cropped
        if not self.is_cropped[self._meas_name]:
            self.crop_signal()

        # Get parameters
        window = "hann"
        nperseg = self.cfg.settings.spectro_nperseg
        detrend = "linear"
        noverlap = self.cfg.settings.spectro_noverlap
        nfft = self.cfg.settings.spectro_win_size

        if noverlap == -1:
            noverlap = nperseg // 2

        # Perform spectrogram
        logger.info(
            "Computing spectrogram (time-window size : "
            f"{nperseg / self.metadata['fs_signal'] * 1e6:2.2f}µs)..."
        )
        fxx, txx, sxx = signal.spectrogram(
            self.get_data_raw(self._meas_name),
            self.metadata["fs_signal"],
            window=window,
            nperseg=nperseg,
            noverlap=noverlap,
            nfft=nfft,
            detrend=detrend,
        )
        logger.info("Spectrogram computed")

        # Store data
        self.set_data_processed("spectro_f", fxx)
        self.set_data_processed("spectro_t", txx)
        self.set_data_processed("spectro_s", sxx)

        return self

    def find_barycenters(self) -> Self:
        """
        Find barycenters in the spectrogram.

        The maxima are found in the 2D PSD. The barycenter of the corresponding
        frequencies is computed in a time window around the maxima, defined by the
        settings in the configuration.

        This results in a time vector (the time windows) and the TDO signal (the
        barycenters, e.g. frequencies). Those two quantities are stored in
        `data_processed`.
        """
        # Checks
        if not self._check_spectro_computed():
            logger.error("Spectrogram not computed. Run `compute_spectrogram()` first")
            return self

        # Get parameters to log
        fxx = self.get_data_processed("spectro_f")
        freq_window = self.cfg.settings.barycenters_fwindow
        nsamples = int(freq_window / np.mean(np.diff(fxx)))
        logger.info(
            "Finding spectrogram barycenters (frequency-window half-size : "
            f"{freq_window:2.3f}Hz, {nsamples} samples)..."
        )
        # Get barycenters
        self.set_data_processed(
            self._tdo_name,
            sp.find_barycenters(
                fxx,
                self.get_data_processed("spectro_s"),
                freq_window,
                fast=self.cfg.settings.barycenters_fast,
            ),
        )
        # Corresponding time vector
        self.apply_time_offset()

        logger.info("TDO signal extracted")

        return self

    def apply_time_offset(self) -> Self:
        """Apply offset on the experiment time vector."""
        if not self.get_data_processed("spectro_t", checkonly=True):
            logger.warning("Spectrogram was not computed")
            return self

        logger.info("Applying offset on time vector...")
        time_tdo = self.get_data_processed("spectro_t")
        self.set_data_processed("time_exp", time_tdo - self.cfg.settings.time_offset)
        self.align_field()
        logger.info(
            f"Applied a {self.cfg.settings.time_offset * 1e6}µs PU/TDO time offset"
        )

        return self

    def detrend_polyfit(self) -> Self:
        """Fit the signal to detrend it."""
        # Checks
        if not self._check_field_aligned():
            self.align_field()
            if not self._check_field_aligned():
                logger.error(
                    "Magnetic field could not be aligned, check messages above"
                )
                return self

        if not hasattr(self, "inds_b_inc") or not hasattr(self, "inds_b_dec"):
            self.get_up_down_indices()

        # Get parameters
        b1, b2 = self.cfg.settings.poly_window
        deg = self.cfg.settings.poly_deg

        # Checks
        maxfield = self.get_data_processed("magfield").max()
        if b1 > maxfield:
            logger.error(
                f"Field window lower bound is {b1}T but max field is {maxfield}T"
            )
            return self

        # Fit and detrend
        logger.info(
            f"Fitting (degree : {deg}) and detrending TDO signal in {b1, b2}T..."
        )
        fp_inc, res_inc = sp.fit_poly(
            self.get_data_processed("magfield")[self.inds_b_inc],
            self.get_data_processed(self._tdo_name)[self.inds_b_inc],
            deg,
            b1,
            b2,
        )

        fp_dec, res_dec = sp.fit_poly(
            self.get_data_processed("magfield")[self.inds_b_dec],
            self.get_data_processed(self._tdo_name)[self.inds_b_dec],
            deg,
            b1,
            b2,
        )

        # Store
        self.set_data_processed(self._tdo_name + "_fit_inc", fp_inc)
        self.set_data_processed(self._tdo_name + "_fit_dec", fp_dec)
        self.set_data_processed(self._tdo_det_inc_name, res_inc)
        self.set_data_processed(self._tdo_det_dec_name, res_dec)

        logger.info("TDO signal detrended")

        return self

    def oversample_inverse_field(self) -> Self:
        """Express data in 1/B and oversample."""
        # Checks
        if not self._check_tdo_detrended():
            self.detrend_polyfit()
            if not self._check_tdo_detrended():
                logger.error("TDO signal could not be detrended")
                return self

        # Get parameters
        b1, b2 = self.cfg.settings.fft_window
        npoints = self.cfg.settings.npoints_interp_inverse

        if b1 == -1:
            b1 = self.cfg.settings.poly_window[0]
        if b2 == -1:
            b2 = self.cfg.settings.poly_window[1]
        if b1 == 0 or b2 == 0:
            logger.error("Can't use 0T as a boundary for FFT (we need 1/B)")
            return self

        # Interpolate inverse field
        logger.info(
            f"Oversampling TDO signal in 1/B ({npoints} points in {b1, b2}T)..."
        )
        field_inverse_inc, tdo_inverse_inc = sp.interpolate_inverse(
            self.get_data_processed("magfield")[self.inds_b_inc],
            self.get_data_processed(self._tdo_det_inc_name),
            b1,
            b2,
            npoints,
        )

        field_inverse_dec, tdo_inverse_dec = sp.interpolate_inverse(
            self.get_data_processed("magfield")[self.inds_b_dec],
            self.get_data_processed(self._tdo_det_dec_name),
            b1,
            b2,
            npoints,
        )

        # Store
        self.set_data_processed("magfield_inverse_inc", field_inverse_inc)
        self.set_data_processed(self._tdo_inv_inc_name, tdo_inverse_inc)
        self.set_data_processed("magfield_inverse_dec", field_inverse_dec)
        self.set_data_processed(self._tdo_inv_dec_name, tdo_inverse_dec)

        logger.info("TDO signal interpolated in 1/B")

        return self

    def fft_inverse_field(self) -> Self:
        """
        FFT on the signal in 1/B.

        The resulting "frequencies" are therefore in units of B (teslas).
        """
        if not self._check_tdo_inverse():
            self.oversample_inverse_field()
            if not self._check_tdo_inverse():
                logger.error("Could not compute FFT because there is no signal in 1/B")
                return self

        # FFT
        logger.info("FFT on TDO signal in 1/B...")
        f_inc, fft_inc = sp.fourier_transform(
            self.get_data_processed(self._tdo_inv_inc_name),
            np.mean(np.diff(self.get_data_processed("magfield_inverse_inc"))),
            pad_mult=self.cfg.settings.fft_pad_mult,
        )

        f_dec, fft_dec = sp.fourier_transform(
            self.get_data_processed(self._tdo_inv_dec_name),
            np.mean(np.diff(self.get_data_processed("magfield_inverse_dec"))),
            pad_mult=self.cfg.settings.fft_pad_mult,
        )

        # Clip
        idx_max_bfreq = sp.find_nearest_index(f_dec, self.cfg.settings.max_bfreq)
        f_inc = f_inc[:idx_max_bfreq]
        fft_inc = fft_inc[:idx_max_bfreq]
        f_dec = f_dec[:idx_max_bfreq]
        fft_dec = fft_dec[:idx_max_bfreq]

        # Store
        self.set_data_processed("bfreq_inc", f_inc)
        self.set_data_processed("fft_inc", fft_inc)
        self.set_data_processed("bfreq_dec", f_dec)
        self.set_data_processed("fft_dec", fft_dec)

        logger.info(f"FFT computed and cropped up until {self.cfg.settings.max_bfreq}T")

        return self

    def extract_tdo(self) -> Self:
        """Extract the TDO signal."""
        self.compute_spectrogram().find_barycenters()

        return self

    def collect_arrays(self, array_names_list: list[str]) -> list[np.ndarray]:
        """Collect and data in `data_processed`."""
        return [self.get_data_processed(name) for name in array_names_list]

    def get_csv_filename(self, suffix: str = "", ext: str = ".csv") -> str:
        """Build the output file name."""
        return str(self.data_directory / (self.expid + suffix + ext))

    def _crop_signal(
        self,
        meas_name: str,
        start: int,
        stop: int,
        where: Literal["raw", "processed"] = "raw",
        force: bool = False,
        append: str = "_full",
    ) -> None:
        """
        Crop a 1D vector from `start` to `stop`.

        The dataset named `meas_name` must be a 1D vector.

        Parameters
        ----------
        meas_name : str
            Dataset name. Must be allowed to be re-cropped, e.g. exists in the
            `is_cropped` dict.
        start, stop : int
            Start and stop index.
        where : {"raw", "processed"}, optional
            Where to get and set the data (`data_raw` or `data_processed`). Default is
            "raw".
        force : bool, optional
            Whether to crop even if a dataset named `meas_name + append` already exists.
            Default is False.
        append : str, optional
            Bit of text that is appended to the base measurement name to store the
            original data. Default is "_full".
        """
        match where:
            case "raw":
                getter = getattr(self, "get_data_raw")
                setter = getattr(self, "set_data_raw")
            case "processed":
                getter = getattr(self, "get_data_processed")
                setter = getattr(self, "set_data_processed")

        if getter(meas_name + append, checkonly=True) and not force:
            self.is_cropped[meas_name] = True
            logger.info(f"{meas_name} not cropped because already cropped")
        elif getter(meas_name, checkonly=True):
            sig = getter(meas_name)
            setter(meas_name + append, sig.copy())
            setter(meas_name, sig[start:stop])
            self.is_cropped[meas_name] = True
            logger.info(f"{meas_name} cropped.")
        else:
            self.is_cropped[meas_name] = False
            logger.warning(f"{meas_name} not cropped because not loaded or created")

    def _remove_data_changed(self) -> None:
        """Remove data that was computed but the data it relied on changed."""
        self.remove_data_processed(self._tdo_name + "_fit_inc")
        self.remove_data_processed(self._tdo_name + "_fit_dec")
        self.remove_data_processed(self._tdo_det_inc_name)
        self.remove_data_processed(self._tdo_det_dec_name)
        self.remove_data_processed(self._tdo_inv_inc_name)
        self.remove_data_processed(self._tdo_inv_dec_name)
        self.remove_data_processed("bfreq_inc")
        self.remove_data_processed("bfreq_dec")
        self.remove_data_processed("fft_inc")
        self.remove_data_processed("fft_dec")

    def _get_result_names_list(self, ud: Literal["inc", "dec"]) -> list[str]:
        """
        Generate the list of things to save.

        Either in increasing or decreasing magnetic field.
        """
        return [
            "time_exp",
            "magfield",
            self._tdo_name,
            self._tdo_name + "_fit_" + ud,
            self._tdo_det_dec_name if ud == "dec" else self._tdo_det_inc_name,
            "magfield_inverse_" + ud,
            self._tdo_inv_dec_name if ud == "dec" else self._tdo_inv_inc_name,
            "bfreq_" + ud,
            "fft_" + ud,
        ]

    def _get_results_up_down(
        self, ud: Literal["inc", "dec"]
    ) -> tuple[np.ndarray, list[str]]:
        """Save increasing or decreasing magnetic field results in a CSV file."""
        names_list = self._get_result_names_list(ud)
        to_save = self.collect_arrays(names_list)
        names_list.insert(3, "magfield_" + ud)
        to_save.insert(
            3,
            self.get_data_processed("magfield")[
                self.inds_b_dec if ud == "dec" else self.inds_b_inc
            ],
        )
        names_list.insert(4, self._tdo_name + "_" + ud)
        to_save.insert(
            4,
            self.get_data_processed(self._tdo_name)[
                self.inds_b_dec if ud == "dec" else self.inds_b_inc
            ],
        )
        return sp.collate_arrays(to_save), names_list

    def _determine_file_format(
        self, filepath: Path, sep: str | None = None
    ) -> tuple[Callable[[Path, str], bool] | str, str]:
        """Determine what type of file we got."""
        match filepath.suffix:
            case ".out":
                return self._load_csv_legacy_tdo, " " if sep is None else sep
            case ".txt":
                return self._load_csv_legacy_results, "," if sep is None else sep
            case ".csv" | ".tsv":
                return self._determine_csv_format(filepath, sep=sep)
            case _:
                logger.error(f"Unknown file extension : {filepath.suffix}")
                return "unknown", ""

    def _determine_csv_format(
        self, filepath: Path, sep: str | None = None
    ) -> tuple[Callable[[Path, str], bool] | str, str]:
        """Determine what type of CSV file we got."""
        # Read first line
        with open(filepath) as f:
            header = f.readline()

        # Count how much delimiters there are to find the number of columns
        if sep is None:
            sep = "\t"
        ncols = header.count(sep) + 1

        if ncols == CSV_TDO_NCOLS:
            return self._load_csv_tdo, sep
        elif ncols == CSV_RES_NCOLS:
            return self._load_csv_results, sep
        else:
            logger.error(
                f"Could not determine what type of file this is : {filepath.name}"
            )
            return "unknown", ""

    def _load_csv_legacy_tdo(self, filepath: Path, sep: str = " ") -> bool:
        """Load already extracted TDO signal from a legacy .out file."""
        logger.info("Loading TDO signal from a (legacy) .out file...")
        data = np.loadtxt(filepath, delimiter=sep, skiprows=1, usecols=(0, 1))

        # data is ["magfield", self._tdo_name]
        self.set_data_processed("magfield", data[:, 0])
        self.set_data_processed(self._tdo_name, data[:, 1])

        # Create the time vector to pass the field-aligned checks
        self.set_data_processed(
            "time_exp", np.linspace(0, self.cfg.settings.max_time, data.shape[0])
        )
        self.set_data_processed("magfield_time", self.get_data_processed("time_exp"))

        return True

    def _load_csv_legacy_results(self, filepath: Path, sep: str = ",") -> bool:
        """Load already computed results from a legacy .txt file."""
        logger.info("Loading results from a (legacy) .txt file...")

        # Load data
        data = np.loadtxt(filepath, delimiter=sep, skiprows=1)

        # Guess if it's increasing or decreasing magnetic field
        inc_or_dec = self._determine_inc_or_dec(filepath)
        if inc_or_dec == "unknown":
            logger.error(
                "Couldn't determine if the file is for increasing or decreasing "
                "magnetic field"
            )
            return False
        self.set_map_in_data_processed(self._map_cols_legacy_results(data, inc_or_dec))

        # Set extra data (to pass field-aligned check)
        self.set_data_processed("magfield_time", self.get_data_processed("time_exp"))

        return True

    def _load_csv_tdo(self, filepath: Path, sep: str = "\t") -> bool:
        """Load already extracted TDO signal from a CSV file."""
        logger.info("Loading TDO signal from CSV file...")
        data = np.loadtxt(filepath, delimiter=sep, skiprows=1, usecols=(0, 1, 2))

        # data is ["time_exp", "magfield", self._tdo_name]
        self.set_data_processed("time_exp", data[:, 0])
        self.set_data_processed("magfield_time", data[:, 0])
        self.set_data_processed("magfield", data[:, 1])
        self.set_data_processed(self._tdo_name, data[:, 2])

        return True

    def _load_csv_results(self, filepath: Path, sep: str = "\t") -> bool:
        """Load already computed results from a CSV file."""
        logger.info("Loading results from a CSV file...")

        # Load data
        data = np.loadtxt(filepath, delimiter=sep, skiprows=1)

        # Guess if it's increasing or decreasing magnetic field
        inc_or_dec = self._determine_inc_or_dec(filepath)
        if inc_or_dec == "unknown":
            logger.error(
                "Couldn't determine if the file is for increasing or decreasing "
                "magnetic field"
            )
            return False
        self.set_map_in_data_processed(self._map_cols_results(data, inc_or_dec))

        # Set extra data (to pass field-aligned check)
        self.set_data_processed("magfield_time", self.get_data_processed("time_exp"))

        return True

    def _determine_inc_or_dec(
        self, filepath: Path, keywords: tuple[str, str] = ("inc", "dec")
    ) -> Literal["inc", "dec", "unknown"]:
        """
        Determine if CSV file is for increasing or decreasing magnetic field.

        Parameters
        ----------
        filepath : Path
            Path to the CSV file.
        keywords : tuple[str]
            Two-elements tuple with the strings that allow to determine if it is
            increasing or decreasing, respectively. Default is ("inc", "dec").

        Returns
        -------
        inc_or_dec : {"inc", "dec", "unknown"}
        """
        with open(filepath) as f:
            header = f.readline()

        ninc = header.count(keywords[0])
        ndec = header.count(keywords[1])

        if ninc > ndec:
            return "inc"
        elif ninc < ndec:
            return "dec"
        else:
            return "unknown"

    def _map_cols_legacy_results(
        self, data: np.ndarray, kw: Literal["inc", "dec"]
    ) -> dict[str, np.ndarray]:
        """
        Map a field name with a column in the legacy results CSV file.

        Parameters
        ----------
        data : np.ndarray
            Tabular data read from the legacy results CSV file.
        kw : {"inc", "dec"}
            Increasing or decreasing magnetic field identifier (usually "inc" or "dec").
        """
        where_nan = np.isnan(data)

        def extract_col(idx):
            return data[~where_nan[:, idx], idx]

        return {
            "time_exp": extract_col(0),
            "magfield": extract_col(1),
            self._tdo_name: extract_col(2),
            self._tdo_name + "_fit_" + kw: extract_col(7),
            self._tdo_det_dec_name
            if kw == "dec"
            else self._tdo_det_inc_name: extract_col(8),
            "magfield_inverse_" + kw: extract_col(9),
            self._tdo_inv_dec_name
            if kw == "dec"
            else self._tdo_inv_inc_name: extract_col(10),
            "bfreq_" + kw: extract_col(11),
            "fft_" + kw: extract_col(12),
        }

    def _map_cols_results(
        self, data: np.ndarray, kw: Literal["inc", "dec"]
    ) -> dict[str, np.ndarray]:
        """
        Map a field name with a column in the results CSV file.

        Parameters
        ----------
        data : np.ndarray
            Tabular data read from the results CSV file.
        kw : {"inc", "dec"}
            Increasing or decreasing magnetic field identifier (usually "inc" or "dec").
        """
        where_nan = np.isnan(data)

        def extract_col(idx):
            return data[~where_nan[:, idx], idx]

        return {
            "time_exp": extract_col(0),
            "magfield": extract_col(1),
            self._tdo_name: extract_col(2),
            self._tdo_name + "_fit_" + kw: extract_col(5),
            self._tdo_det_dec_name
            if kw == "dec"
            else self._tdo_det_inc_name: extract_col(6),
            "magfield_inverse_" + kw: extract_col(7),
            self._tdo_inv_dec_name
            if kw == "dec"
            else self._tdo_inv_inc_name: extract_col(8),
            "bfreq_" + kw: extract_col(9),
            "fft_" + kw: extract_col(10),
        }

    def _check_pickup_loaded(self) -> bool:
        """Check if the pickup data was loaded."""
        if ("pickup" not in self.data_raw) or ("fs_pickup" not in self.metadata):
            return False
        else:
            return True

    def _check_data_loaded(self) -> bool:
        """Check if oscilloscope data was loaded."""
        return (self._meas_name in self.data_raw) and ("fs_signal" in self.metadata)

    def _check_spectro_computed(self) -> bool:
        """Check if the spectrogram was computed."""
        return self.get_data_processed("spectro_s", checkonly=True)

    def _check_barycenters_computed(self) -> bool:
        """Check if the frequencies barycenters were computed."""
        return self.get_data_processed(
            self._tdo_name, checkonly=True
        ) and self.get_data_processed("time_exp", checkonly=True)

    def _check_tdo_detrended(self) -> bool:
        """Check if the TDO signal has been detrended."""
        return self.get_data_processed(
            self._tdo_det_inc_name, checkonly=True
        ) and self.get_data_processed(self._tdo_det_dec_name, checkonly=True)

    def _check_tdo_inverse(self) -> bool:
        """Check if the TDO signal was oversampled in 1/B."""
        return self.get_data_processed(
            self._tdo_inv_inc_name, checkonly=True
        ) and self.get_data_processed(self._tdo_inv_dec_name, checkonly=True)



"""Command line interface to run the various apps."""

import importlib.metadata
from typing import Annotated

import typer

VERSION = importlib.metadata.version("pymagnetos")


app = typer.Typer()


@app.command()
def pyuson():
    """Run the app for ultrasound experiments."""
    from pymagnetos.pyuson import gui

    gui.run()


@app.command()
def pytdo():
    """Run the app for TDO experiments."""
    from pymagnetos.pytdo import gui

    gui.run()


@app.command()
def linx():
    """Run the app for digital lock-in."""
    from pymagnetos.linx import gui

    gui.run()


def version_callback(value: bool):
    """Print version."""
    if value:
        print(f"pymagnetos version {VERSION}")
        raise typer.Exit()


@app.callback()
def main(
    version: Annotated[
        bool | None,
        typer.Option(
            "--version", callback=version_callback, help="Show version number and exit."
        ),
    ] = None,
):
    """Handle global options."""
    pass


if __name__ == "__main__":
    app()



"""
The pymagnetos package.

It provides analysis tools for high magnetic field experiments performed at LNCMI-T.

Experiment-specific library are specialized implementations of the base classes found in
the `core` subpackage. They should be generic enough to be used to build any pipeline
relying on a 'configuration > analysis > store' workflow :

1. Configuration : handled by the `Config` object, which uses `pydantic` models to read
and validate TOML configuration files. One needs to define the `pydantic` models. The
`BaseConfig` object provides the TOML parser and writer and interface with `pydantic`.
2. Analysis : handled by the `Processor` objects. It is initialized with the path to a
configuration file. The `BaseProcessor` object provides the base framework and some
shared methods related to magnetic field computation.
3. Store : the `Processor` objects uses the `nexusformat` library to store data
in-memory. The whole data structure can be easily dumped to a NeXus-compliant HDF5 file,
and can be consolidated (i.e. add sample- and experiment-related metadata) via the
configuration file. The output NeXus file is self-contained : it has the raw and
processed data, along with the analysis configuration. The `Processor` objects can be
initialized with such a file instead of a configuration file.

Graphical user interface
------------------------
pymagnetos provides graphical applications built with PyQt6. Each subpackages have a
`gui` submodule (that are not imported by default). The `core.gui` module contains the
base components to build an app with a configuration panel, a zone with graphs, action
buttons and a log text box. It interfaces with a `Worker` object, which is a `Processor`
instance set in its own thread. The components are Qt widgets that can be built upon and
re-used.

Available subpackages
---------------------
- `core` : base classes and components.
- `pytdo` : a library for TDO experiments.
- `pyuson` : a library for ultrasound experiments.

Available modules
-----------------
- `sp` : signal processing functions.
- `utils` : utilities for dictionary manipulation.
- `cli` : used to run the applications.
"""

import nexusformat.nexus as nx

from . import core, linx, pytdo, pyuson
from .core import sp, utils

__all__ = ["core", "linx", "pytdo", "pyuson", "sp", "utils"]

# Configure NeXus globally
nx.nxsetconfig(compression=None, encoding="utf-8", lock=0, memory=8000, recursive=True)



"""Utility functions."""

from collections.abc import Mapping
from typing import Any


def merge_dict_nested(
    dic_user: Mapping[str, Any], dic_default: dict[str, Any]
) -> dict[str, Any]:
    """
    Merge two dictionaries.

    Keep elements from `dic_user` if it exists, with a fallback on elements in
    `dic_default`. All nested dictionaries are also treated like this.

    Parameters
    ----------
    dic_user : dict-like
        Dict being merged, its values have priority over `dic_default`.
    dic_default : dict-like
        Dict being merged, its values are used if they are not in `dic_user`.

    Returns
    -------
    dic : dict
        Merged dict.
    """
    dic_copy = dic_default.copy()
    for key, value in dic_user.items():
        if key in dic_default:
            if isinstance(value, Mapping):
                dic_copy[key] = merge_dict_nested(value, dic_default[key])
            else:
                if key in dic_user:
                    dic_copy[key] = dic_user[key]

        else:
            dic_copy[key] = dic_user[key]

    return dic_copy


def cast_type_dict(
    dic: dict[str, Any], in_type: type, out_type: type
) -> dict[str, Any]:
    """
    Convert `in_type` value to `out_type` values in a dictionary.

    Nested dictionaries are supported, so the casted types can't be `dict`.

    Parameters
    ----------
    dic : dict
        Dict in which values are casted.
    in_type, out_type : type
        `in_type` types are casted to `out_type`. Can't be `dict`.
    """
    dic_copy = dic.copy()
    for key, value in dic.items():
        if isinstance(value, dict):
            dic_copy[key] = cast_type_dict(value, in_type, out_type)
        else:
            if isinstance(dic_copy[key], in_type):
                dic_copy[key] = out_type(dic[key])

    return dic_copy


def strip_none_dict(dic: dict) -> dict:
    """
    Strip elements that are None from `dic`.

    Parameters
    ----------
    dic : dict
        Dict in which None values are removed.

    Returns
    -------
    dic_stripped : dict
        Input dictionary where None values were removed.
    """
    dic_copy = dic.copy()
    for key, value in dic.items():
        if isinstance(value, dict):
            dic_copy[key] = strip_none_dict(value)
        elif value is None:
            dic_copy.pop(key)

    return dic_copy


def remove_duplicate_values(
    dic1: dict[str, Any], dic2: dict[str, Any]
) -> dict[str, Any]:
    """
    Remove entries with duplicate value from `dic1` if it is in `dic2`.

    The entry in `dic1` is only removed if it has another entry with the same value
    (whatever its key).

    Parameters
    ----------
    dic1 : dict
        Dictionary in which entries with duplicate value are removed.
    dic2 : dict
        Dictionary checked against `dic1`.

    Returns
    -------
    dic : dict
        Same as `dic1`, with entries whose value are found in `dic2` were removed.
    """
    # Reverse dict to check duplicate values
    dic1_reverse = dict()
    for key, value in dic1.items():
        # Supply an empty list for missing key
        dic1_reverse.setdefault(value, list()).append(key)

    # List keys that have duplicate values
    duplicates = flatten_list(
        [values for key, values in dic1_reverse.items() if len(values) > 1]
    )

    # Remove entries with duplicate values that are in dic2
    dic1_copy = dic1.copy()
    for key in duplicates:
        if key in dic2 and dic1[key] == dic2[key]:
            dic1_copy.pop(key)

    return dic1_copy


def flatten_list(list_of_lists: list[list[Any]]) -> list[Any]:
    """
    Flatten a list of lists.

    Parameters
    ----------
    list_of_lists : list of lists
        The list of lists to flatten.

    Returns
    -------
    flat_list : list
        Flat list.
    """
    flat_list = []
    for lst in list_of_lists:
        flat_list += lst
    return flat_list



"""The grid with main action buttons."""

from PyQt6 import QtWidgets
from PyQt6.QtCore import pyqtSignal


class ButtonsWidget(QtWidgets.QWidget):
    sig_load = pyqtSignal()
    sig_find_f0 = pyqtSignal()
    sig_auto_phase = pyqtSignal()
    sig_export_csv = pyqtSignal()
    sig_save_nexus = pyqtSignal()

    def __init__(self) -> None:
        super().__init__()

        # Create buttons
        self.button_load = QtWidgets.QPushButton("Load data", self)
        self.button_find_f0 = QtWidgets.QPushButton("Find f0", self)
        self.button_auto_phase = QtWidgets.QPushButton("Auto-phase", self)
        self.button_export_csv = QtWidgets.QPushButton("Export as CSV", self)
        self.button_save_nexus = QtWidgets.QPushButton("Save as NeXus", self)

        export_options_widget = self.create_export_options()

        self.disable_buttons()

        self.connect_buttons()

        grid = QtWidgets.QGridLayout()
        grid.addWidget(self.button_load, 0, 0, 1, 2)
        grid.addWidget(self.button_find_f0, 1, 0, 1, 1)
        grid.addWidget(self.button_auto_phase, 1, 1, 1, 1)
        grid.addWidget(self.button_export_csv, 2, 0, 1, 1)
        grid.addWidget(self.button_save_nexus, 2, 1, 1, 1)
        grid.addLayout(export_options_widget, 3, 0, 1, 2)

        self.setLayout(grid)

    @property
    def downsample_factor(self) -> int:
        return self.spinbox_downsample.value()

    @property
    def downsample_averaging(self) -> bool:
        return self.checkbox_downsample.isChecked()

    def create_export_options(self) -> QtWidgets.QHBoxLayout:
        """Create the layout with export options."""
        downsample_label = QtWidgets.QLabel("Downsampling x")
        self.spinbox_downsample = QtWidgets.QSpinBox()

        downsample_widget = QtWidgets.QWidget(self)
        downsample_grid = QtWidgets.QHBoxLayout()
        downsample_grid.addWidget(downsample_label)
        downsample_grid.addWidget(self.spinbox_downsample)
        downsample_widget.setLayout(downsample_grid)

        self.checkbox_downsample = QtWidgets.QCheckBox("Averaging")

        grid = QtWidgets.QHBoxLayout()
        grid.addWidget(downsample_widget)
        grid.addWidget(self.checkbox_downsample)

        return grid

    def connect_buttons(self) -> None:
        self.button_load.clicked.connect(self.sig_load.emit)
        self.button_find_f0.clicked.connect(self.sig_find_f0.emit)
        self.button_auto_phase.clicked.connect(self.sig_auto_phase)
        self.button_export_csv.clicked.connect(self.sig_export_csv.emit)
        self.button_save_nexus.clicked.connect(self.sig_save_nexus.emit)

    def disable_buttons(self) -> None:
        self.button_load.setEnabled(False)
        self.button_find_f0.setEnabled(False)
        self.button_auto_phase.setEnabled(False)
        self.button_export_csv.setEnabled(False)
        self.button_save_nexus.setEnabled(False)

    def enable_buttons(self) -> None:
        self.button_load.setEnabled(False)
        self.button_find_f0.setEnabled(True)
        self.button_auto_phase.setEnabled(True)
        self.button_export_csv.setEnabled(True)
        self.button_save_nexus.setEnabled(True)



"""
The core subpackage, containing base classes and components.

Those are meant to be subclassed to build experiment-specific objects.

Provides
--------
- `BaseConfig` : the base class for the `Config` object.
- `BaseProcessor` : the base class for the `Processor` object.
- `BaseData` : NeXus base interface.
- `DataRaw` : NeXus interface for a raw data container.
- `DataProcessed` : NeXus interface for a processed data container.
- `config_models` : `pydantic` models to define the configuration.
- `sp` : signal processing functions.
- `utils` : dictionary manipulation utility functions.
"""

from . import config_models, utils
from . import signal_processing as sp
from ._baseprocessor import BaseProcessor
from ._config import BaseConfig
from ._data import BaseData, DataProcessed, DataRaw

__all__ = [
    "BaseConfig",
    "BaseProcessor",
    "BaseData",
    "DataProcessed",
    "DataRaw",
    "config_models",
    "sp",
    "utils",
]



"""Pydantic models for the TDOProcessor Config."""

from collections.abc import Sequence
from pathlib import Path
from typing import Any

from pydantic import BaseModel, ConfigDict

from ..core.config_models import File, Nexus, Parameters


class Settings(BaseModel):
    """Settings for signal extraction, analysis and plotting."""

    # Extraction
    max_time: float

    spectro_nperseg: int
    spectro_win_size: int
    spectro_noverlap: int

    barycenters_fwindow: float
    barycenters_fast: bool

    time_offset: float

    # Analysis
    poly_window: Sequence[float]
    poly_deg: int
    npoints_interp_inverse: int
    fft_window: Sequence[float]
    fft_pad_mult: int
    max_bfreq: float

    # Display
    offset: float


class TDOConfigurationModel(BaseModel):
    """A model specialized for TDO experiments."""

    expid: str
    data_directory: Path

    filenames: dict[str, str | Path] = dict()

    files: dict[str, File]

    measurements: dict[str, int]

    parameters: Parameters
    settings: Settings

    nx: dict[str, dict[str, Any]]
    nexus: Nexus

    model_config = ConfigDict(validate_assignment=True)



"""The graph area widget for Linx."""

import pyqtgraph as pg
from PyQt6 import QtCore, QtWidgets

from pymagnetos.core.gui.widgets import BaseGraphsWidget


class GraphsWidget(BaseGraphsWidget):
    """The graphs area with all the plots."""

    def __init__(self) -> None:
        super().__init__()

        # Create empty canvases in tabs
        field_tab = self.create_field_plot()
        signal_tab = self.create_signal_plot()
        res_time_tab = self.create_sigtime_plot()
        res_field_tab = self.create_sigfield_plot()

        self.init_coordinates_on_hover()
        self.disable_autoSI_prefix()

        # Create a main splitter for rows (vertical)
        main_splitter = QtWidgets.QSplitter(QtCore.Qt.Orientation.Vertical)

        # Create a top splitter for the first row (horizontal)
        top_splitter = QtWidgets.QSplitter(QtCore.Qt.Orientation.Horizontal)
        top_splitter.addWidget(field_tab)
        top_splitter.addWidget(signal_tab)
        top_splitter.setStretchFactor(0, 1)  # field_tab stretch factor
        # frame_tab spans 3 columns
        top_splitter.setStretchFactor(1, 3)

        # Create a bottom splitter for the second row (horizontal)
        bottom_splitter = QtWidgets.QSplitter(QtCore.Qt.Orientation.Horizontal)
        bottom_splitter.addWidget(res_time_tab)
        bottom_splitter.addWidget(res_field_tab)
        # res_time_tab spans 2 columns
        bottom_splitter.setStretchFactor(0, 2)
        # res_field_tab spans 2 columns
        bottom_splitter.setStretchFactor(1, 2)

        # Add the top and bottom splitters to the main splitter
        main_splitter.addWidget(top_splitter)
        main_splitter.addWidget(bottom_splitter)
        main_splitter.setStretchFactor(0, 1)  # top row stretch factor
        main_splitter.setStretchFactor(1, 1)  # bottom row stretch factor

        # Set the main splitter as the central widget
        layout = QtWidgets.QVBoxLayout()
        layout.addWidget(main_splitter)
        self.setLayout(layout)

    def create_field_plot(self) -> QtWidgets.QTabWidget:
        tab = QtWidgets.QTabWidget(self)

        # magnetic field (integrated)
        self.field = pg.PlotWidget(title="Magnetic field")
        self.field.setLabel("bottom", "time (s)")
        self.field.setLabel("top", "frame (#) - field not aligned")
        self.field.setLabel("left", "field (T)")
        self.field.showGrid(y=True)
        self._plots_list.add(self.field)

        # pickup coil voltage (measured)
        self.dfield = pg.PlotWidget(title="Pickup")
        self.dfield.setLabel("bottom", "time (s)")
        self.dfield.setLabel("top", "frame (#)")
        self.dfield.setLabel("left", "pickup (V)")
        self.dfield.showGrid(y=True)
        self._plots_list.add(self.dfield)

        tab.addTab(self.field, "B(t)")
        tab.addTab(self.dfield, "Pickup")

        return tab

    def create_signal_plot(self) -> QtWidgets.QTabWidget:
        tab = QtWidgets.QTabWidget(self)

        # Raw signal
        self.raw_signal = pg.PlotWidget(title="Raw signal")
        self.raw_signal.setLabel("bottom", "time (s)")
        self.raw_signal.setLabel("left", "signal (V)")
        self.raw_signal.showGrid(y=True)
        self._plots_list.add(self.raw_signal)

        # Reference signal
        self.reference = pg.PlotWidget(title="Reference signal")
        self.reference.setLabel("bottom", "time (s)")
        self.reference.setLabel("left", "signal (V)")
        self.reference.showGrid(y=True)
        self._plots_list.add(self.reference)

        tab.addTab(self.raw_signal, "Raw signal")
        tab.addTab(self.reference, "Reference signal")

        return tab

    def create_sigtime_plot(self) -> QtWidgets.QTabWidget:
        tab = QtWidgets.QTabWidget(self)

        # vs time
        self.res_time = pg.PlotWidget(title="Resistance (t)")
        self.res_time.setLabel("bottom", "time (s)")
        self.res_time.setLabel("left", "resistance (ohm)")
        self.res_time.showGrid(y=True)
        self._plots_list.add(self.res_time)

        tab.addTab(self.res_time, "Resistance (t)")

        return tab

    def create_sigfield_plot(self) -> QtWidgets.QTabWidget:
        tab = QtWidgets.QTabWidget(self)

        # vs field
        self.res_field = pg.PlotWidget(title="Resistance (B)")
        self.res_field.setLabel("bottom", "field (T)")
        self.res_field.setLabel("left", "resistance (ohm)")
        self.res_field.showGrid(y=True)
        self.res_field.getPlotItem().addLegend()
        self._plots_list.add(self.res_field)

        tab.addTab(self.res_field, "Resistance (B)")

        return tab

    def init_plot_style(self) -> None:
        """Set up PyQtGraph line styles."""
        w0 = 1
        w1 = 1
        self.pen_field = pg.mkPen("#c7c7c7", width=w0)
        self.pen_raw_signal = pg.mkPen("#f9efce", width=w0)
        self.pen_reference = pg.mkPen("#E3F988", width=w0)
        self.pen_in_phase_bdown = pg.mkPen("#00bfff", width=w1)
        self.pen_in_phase_bup = pg.mkPen("#4169e1", width=w1)
        self.pen_out_phase_bdown = pg.mkPen("#ff7f50", width=w1)
        self.pen_out_phase_bup = pg.mkPen("#ff4500", width=w1)

    def enable_rois(self) -> None:
        """Required by core."""
        return
        
    def disable_rois(self) -> None:
        """Required by core."""


"""
The `pytdo` subpackage, a library for TDO experiments.

Provides
--------
- `TDOConfig` : The `Config` object specific to TDO experiments.
- `TDOProcessor` : the `Processor` object specific to TDO experiments.
- `config_models` : the `pydantic` models to define the configuration requirements.
- `gui` : a graphical application. It needs to be imported explicitely.
"""

from . import config_models
from ._config import TDOConfig
from ._tdoprocessor import TDOProcessor

__all__ = ["TDOConfig", "TDOProcessor", "config_models"]



"""Worker that connects to a Processor object."""

from pathlib import Path
from typing import Any

from PyQt6.QtCore import QObject
from PyQt6.QtCore import pyqtSignal as Signal
from PyQt6.QtCore import pyqtSlot as Slot

from pymagnetos.core import BaseProcessor


class BaseWorker(QObject):
    """
    Wrapper around a Processor object.

    It exposes the Processor methods as slots and emits a signal when the task is
    finished. This allows putting the worker in a separate thread to keep the app
    responsive.

    Subclasses must register the type of Processor before calling the Base `__init__()`
    method. It is initialized with any arguments the Processor class accepts.

    Signals
    -------
    sig_load_data_finished : emits when data is loaded.
    sig_compute_field_finished : emits when the magnetic field is computed.
    sig_preprocess_finished : emits when preprocessing steps are finished.
    sig_align_finished : emits when the magnetic field is aligned on the data.
    sig_analysis_finished : emits when the analysis is finished.
    sig_save_nexus_finished : emits when the NeXus file is saved.
    sig_batch_finished : should be used to emit when the batch processing is finished.
    sig_preprocess_progress : should be used to emit step (int) during preprocessing.
    sig_batch_process_progress : should be used to emit step (int) index during
        batch-processing.
    """

    _type_processor: type[BaseProcessor] = BaseProcessor

    # Additional actions to take after loading
    _after_load_preprocess: bool = False
    _after_load_align: bool = False
    _after_load_analysis: bool = False

    # Additional arguments for the load, preprocess, run_analysis methods
    _kwargs_load: dict[str, Any] = {}
    _kwargs_preprocess: dict[str, Any] = {}
    _kwargs_analysis: dict[str, Any] = {}

    proc: BaseProcessor
    is_data_loaded: bool

    sig_load_data_finished: Signal = Signal()
    sig_compute_field_finished: Signal = Signal()
    sig_preprocess_finished: Signal = Signal()
    sig_align_finished: Signal = Signal()
    sig_analysis_finished: Signal = Signal()
    sig_save_nexus_finished: Signal = Signal()
    sig_batch_finished: Signal = Signal()
    sig_preprocess_progress: Signal = Signal(int)
    sig_batch_progress: Signal = Signal(int)

    def __init__(self, *args, **kwargs) -> None:

        super().__init__()

        # Create the Processor object
        self.proc = self._type_processor(*args, **kwargs)
        if self.proc.is_nexus_file:
            self.is_data_loaded = True
        else:
            self.is_data_loaded = False

    @Slot()
    def load_data(self) -> None:
        """
        Load data.

        Optionally run the preprocessing, align the magnetic field and run the analysis.
        **kwargs are passed to the Processor `load_data()` method.
        """
        self.proc.load_data(**self._kwargs_load)
        self.is_data_loaded = True
        if self._after_load_preprocess:
            self.proc.preprocess(**self._kwargs_preprocess)
        if self._after_load_align:
            self.proc.align_field()
        if self._after_load_analysis:
            self.proc.run_analysis(**self._kwargs_analysis)

        self.sig_load_data_finished.emit()

    @Slot()
    def compute_field(self) -> None:
        """Compute magnetic field."""
        self.proc.compute_field()
        self.sig_compute_field_finished.emit()

    @Slot()
    def preprocess(self) -> None:
        """Run the preprocessing steps."""
        self.proc.preprocess(**self._kwargs_preprocess)
        self.sig_preprocess_finished.emit()

    @Slot()
    def align_field(self) -> None:
        """Align magnetic field on data."""
        self.proc.align_field()
        self.sig_align_finished.emit()

    @Slot()
    def run_analysis(self) -> None:
        """Run the analysis."""
        self.proc.run_analysis(**self._kwargs_analysis)
        self.sig_analysis_finished.emit()

    @Slot(str)
    def save_as_nexus(self, fname: str | Path) -> None:
        """Save data as NeXus."""
        self.proc.save(filename=fname)
        self.sig_save_nexus_finished.emit()

    @Slot(tuple)
    def batch_process(self, args: tuple) -> None:
        """Run batch processing."""
        pass
